import numpy as np


def to_center_form(boxes):
    """Transform from corner coordinates to center coordinates.

    # Arguments
        boxes: Numpy array with shape `(num_boxes, 4)`.

    # Returns
        Numpy array with shape `(num_boxes, 4)`.
    """
    x_min, y_min = boxes[:, 0:1], boxes[:, 1:2]
    x_max, y_max = boxes[:, 2:3], boxes[:, 3:4]
    center_x = (x_max + x_min) / 2.0
    center_y = (y_max + y_min) / 2.0
    W = x_max - x_min
    H = y_max - y_min
    return np.concatenate([center_x, center_y, W, H], axis=1)


def to_corner_form(boxes):
    """Transform from center coordinates to corner coordinates.

    # Arguments
        boxes: Numpy array with shape `(num_boxes, 4)`.

    # Returns
        Numpy array with shape `(num_boxes, 4)`.
    """
    center_x, center_y = boxes[:, 0:1], boxes[:, 1:2]
    W, H = boxes[:, 2:3], boxes[:, 3:4]
    x_min = center_x - (W / 2.0)
    x_max = center_x + (W / 2.0)
    y_min = center_y - (H / 2.0)
    y_max = center_y + (H / 2.0)
    return np.concatenate([x_min, y_min, x_max, y_max], axis=1)


def encode(matched, priors, variances=[0.1, 0.1, 0.2, 0.2]):
    """Encode the variances from the priorbox layers into the ground truth
    boxes we have matched (based on jaccard overlap) with the prior boxes.

    # Arguments
        matched: Numpy array of shape `(num_priors, 4)` with boxes in
            point-form.
        priors: Numpy array of shape `(num_priors, 4)` with boxes in
            center-form.
        variances: (list[float]) Variances of priorboxes

    # Returns
        encoded boxes: Numpy array of shape `(num_priors, 4)`.
    """
    boxes = matched[:, :4]
    boxes = to_center_form(boxes)
    center_difference_x = boxes[:, 0:1] - priors[:, 0:1]
    encoded_center_x = center_difference_x / priors[:, 2:3]
    center_difference_y = boxes[:, 1:2] - priors[:, 1:2]
    encoded_center_y = center_difference_y / priors[:, 3:4]
    encoded_center_x = encoded_center_x / variances[0]
    encoded_center_y = encoded_center_y / variances[1]
    encoded_W = np.log((boxes[:, 2:3] / priors[:, 2:3]) + 1e-8)
    encoded_H = np.log((boxes[:, 3:4] / priors[:, 3:4]) + 1e-8)
    encoded_W = encoded_W / variances[2]
    encoded_H = encoded_H / variances[3]
    encoded_boxes = [encoded_center_x, encoded_center_y, encoded_W, encoded_H]
    return np.concatenate(encoded_boxes + [matched[:, 4:]], axis=1)


def decode(predictions, priors, variances=[0.1, 0.1, 0.2, 0.2]):
    """Decode default boxes into the ground truth boxes

    # Arguments
        loc: Numpy array of shape `(num_priors, 4)`.
        priors: Numpy array of shape `(num_priors, 4)`.
        variances: List of two floats. Variances of prior boxes.

    # Returns
        decoded boxes: Numpy array of shape `(num_priors, 4)`.
    """
    center_x = predictions[:, 0:1] * priors[:, 2:3] * variances[0]
    center_x = center_x + priors[:, 0:1]
    center_y = predictions[:, 1:2] * priors[:, 3:4] * variances[1]
    center_y = center_y + priors[:, 1:2]
    W = priors[:, 2:3] * np.exp(predictions[:, 2:3] * variances[2])
    H = priors[:, 3:4] * np.exp(predictions[:, 3:4] * variances[3])
    boxes = np.concatenate([center_x, center_y, W, H], axis=1)
    boxes = to_corner_form(boxes)
    return np.concatenate([boxes, predictions[:, 4:]], 1)


def compute_ious(boxes_A, boxes_B):
    """Calculates the intersection over union between `boxes_A` and `boxes_B`.
    For each box present in the rows of `boxes_A` it calculates
    the intersection over union with respect to all boxes in `boxes_B`.
    The variables `boxes_A` and `boxes_B` contain the corner coordinates
    of the left-top corner `(x_min, y_min)` and the right-bottom
    `(x_max, y_max)` corner.

    # Arguments
        boxes_A: Numpy array with shape `(num_boxes_A, 4)`.
        boxes_B: Numpy array with shape `(num_boxes_B, 4)`.

    # Returns
        Numpy array of shape `(num_boxes_A, num_boxes_B)`.
    """
    xy_min = np.maximum(boxes_A[:, None, 0:2], boxes_B[:, 0:2])
    xy_max = np.minimum(boxes_A[:, None, 2:4], boxes_B[:, 2:4])
    intersection = np.maximum(0.0, xy_max - xy_min)
    intersection_area = intersection[:, :, 0] * intersection[:, :, 1]
    areas_A = (boxes_A[:, 2] - boxes_A[:, 0]) * (boxes_A[:, 3] - boxes_A[:, 1])
    areas_B = (boxes_B[:, 2] - boxes_B[:, 0]) * (boxes_B[:, 3] - boxes_B[:, 1])
    # broadcasting for outer sum i.e. a sum of all possible combinations
    union_area = (areas_A[:, np.newaxis] + areas_B) - intersection_area
    union_area = np.maximum(union_area, 1e-8)
    return np.clip(intersection_area / union_area, 0.0, 1.0)


def compute_max_matches(boxes, prior_boxes):
    iou_matrix = compute_ious(prior_boxes, boxes)
    per_prior_which_box_iou = np.max(iou_matrix, axis=1)
    per_prior_which_box_arg = np.argmax(iou_matrix, axis=1)
    return per_prior_which_box_iou, per_prior_which_box_arg


def get_matches_masks(boxes, prior_boxes, positive_iou=0.5, negative_iou=0.4):
    prior_boxes = to_corner_form(prior_boxes)
    max_matches = compute_max_matches(boxes, prior_boxes)
    per_prior_which_box_iou, per_prior_which_box_arg = max_matches
    positive_mask = np.greater_equal(per_prior_which_box_iou, positive_iou)
    negative_mask = np.less(per_prior_which_box_iou, negative_iou)
    not_ignoring_mask = np.logical_or(positive_mask, negative_mask)
    # ignoring mask are all masks not positive or negative
    ignoring_mask = np.logical_not(not_ignoring_mask)
    return per_prior_which_box_arg, positive_mask, ignoring_mask


def mask_classes(boxes, positive_mask, ignoring_mask):
    class_indices = boxes[:, 4]
    negative_mask = np.not_equal(positive_mask, 1.0)
    class_indices = np.where(negative_mask, 0.0, class_indices)
    # ignoring_mask = np.equal(ignoring_mask, 1.0)
    # class_indices = np.where(ignoring_mask, -1.0, class_indices)
    class_indices = np.expand_dims(class_indices, axis=-1)
    boxes[:, 4:5] = class_indices
    return boxes


def match_beta(boxes, prior_boxes, positive_iou=0.5, negative_iou=0.0):
    """Matches each prior box with a ground truth box (box from `boxes`).
    It then selects which matched box will be considered positive e.g. iou > .5
    and returns for each prior box a ground truth box that is either positive
    (with a class argument different than 0) or negative.

    # Arguments
        boxes: Numpy array of shape `(num_ground_truh_boxes, 4 + 1)`,
            where the first the first four coordinates correspond to
            box coordinates and the last coordinates is the class
            argument. This boxes should be the ground truth boxes.
        prior_boxes: Numpy array of shape `(num_prior_boxes, 4)`.
            where the four coordinates are in center form coordinates.
        positive_iou: Float between [0, 1]. Intersection over union
            used to determine which box is considered a positive box.
        negative_iou: Float between [0, 1]. Intersection over union
            used to determine which box is considered a negative box.

    # Returns
        numpy array of shape `(num_prior_boxes, 4 + 1)`.
            where the first the first four coordinates correspond to point
            form box coordinates and the last coordinates is the class
            argument.
    """
    matches = get_matches_masks(boxes, prior_boxes, positive_iou, negative_iou)
    per_prior_box_which_box_arg, positive_mask, ignoring_mask = matches
    matched_boxes = np.take(boxes, per_prior_box_which_box_arg, axis=0)
    matched_boxes = mask_classes(matched_boxes, positive_mask, ignoring_mask)
    return matched_boxes


def match(boxes, prior_boxes, iou_threshold=0.5):
    """Matches each prior box with a ground truth box (box from `boxes`).
    It then selects which matched box will be considered positive e.g. iou > .5
    and returns for each prior box a ground truth box that is either positive
    (with a class argument different than 0) or negative.

    # Arguments
        boxes: Numpy array of shape `(num_ground_truh_boxes, 4 + 1)`,
            where the first the first four coordinates correspond to
            box coordinates and the last coordinates is the class
            argument. This boxes should be the ground truth boxes.
        prior_boxes: Numpy array of shape `(num_prior_boxes, 4)`.
            where the four coordinates are in center form coordinates.
        iou_threshold: Float between [0, 1]. Intersection over union
            used to determine which box is considered a positive box.

    # Returns
        numpy array of shape `(num_prior_boxes, 4 + 1)`.
            where the first the first four coordinates correspond to point
            form box coordinates and the last coordinates is the class
            argument.
    """
    ious = compute_ious(boxes, to_corner_form(np.float32(prior_boxes)))
    per_prior_which_box_iou = np.max(ious, axis=0)
    per_prior_which_box_arg = np.argmax(ious, 0)

    #  overwriting per_prior_which_box_arg if they are the best prior box
    per_box_which_prior_arg = np.argmax(ious, 1)
    per_prior_which_box_iou[per_box_which_prior_arg] = 2
    for box_arg in range(len(per_box_which_prior_arg)):
        best_prior_box_arg = per_box_which_prior_arg[box_arg]
        per_prior_which_box_arg[best_prior_box_arg] = box_arg

    matches = boxes[per_prior_which_box_arg]
    matches[per_prior_which_box_iou < iou_threshold, 4] = 0
    return matches


def compute_iou(box, boxes):
    """Calculates the intersection over union between 'box' and all 'boxes'.
    Both `box` and `boxes` are in corner coordinates.

    # Arguments
        box: Numpy array with length at least of 4.
        boxes: Numpy array with shape `(num_boxes, 4)`.

    # Returns
        Numpy array of shape `(num_boxes, 1)`.
    """

    x_min_A, y_min_A, x_max_A, y_max_A = box[:4]
    x_min_B, y_min_B = boxes[:, 0], boxes[:, 1]
    x_max_B, y_max_B = boxes[:, 2], boxes[:, 3]
    # calculating the intersection
    inner_x_min = np.maximum(x_min_B, x_min_A)
    inner_y_min = np.maximum(y_min_B, y_min_A)
    inner_x_max = np.minimum(x_max_B, x_max_A)
    inner_y_max = np.minimum(y_max_B, y_max_A)
    inner_w = np.maximum((inner_x_max - inner_x_min), 0)
    inner_h = np.maximum((inner_y_max - inner_y_min), 0)
    intersection_area = inner_w * inner_h
    # calculating the union
    box_area_B = (x_max_B - x_min_B) * (y_max_B - y_min_B)
    box_area_A = (x_max_A - x_min_A) * (y_max_A - y_min_A)
    union_area = box_area_A + box_area_B - intersection_area
    intersection_over_union = intersection_area / union_area
    return intersection_over_union


def apply_non_max_suppression(boxes, scores, iou_thresh=.45, top_k=200):
    """Apply non maximum suppression.

    # Arguments
        boxes: Numpy array, box coordinates of shape `(num_boxes, 4)`
            where each columns corresponds to x_min, y_min, x_max, y_max.
        scores: Numpy array, of scores given for each box in `boxes`.
        iou_thresh: float, intersection over union threshold for removing
            boxes.
        top_k: int, number of maximum objects per class.

    # Returns
        selected_indices: Numpy array, selected indices of kept boxes.
        num_selected_boxes: int, number of selected boxes.
    """

    selected_indices = np.zeros(shape=len(scores))
    if boxes is None or len(boxes) == 0:
        return selected_indices
    x_min = boxes[:, 0]
    y_min = boxes[:, 1]
    x_max = boxes[:, 2]
    y_max = boxes[:, 3]
    areas = (x_max - x_min) * (y_max - y_min)
    remaining_sorted_box_indices = np.argsort(scores)
    remaining_sorted_box_indices = remaining_sorted_box_indices[-top_k:]

    num_selected_boxes = 0
    while len(remaining_sorted_box_indices) > 0:
        best_score_args = remaining_sorted_box_indices[-1]
        selected_indices[num_selected_boxes] = best_score_args
        num_selected_boxes = num_selected_boxes + 1
        if len(remaining_sorted_box_indices) == 1:
            break

        remaining_sorted_box_indices = remaining_sorted_box_indices[:-1]

        best_x_min = x_min[best_score_args]
        best_y_min = y_min[best_score_args]
        best_x_max = x_max[best_score_args]
        best_y_max = y_max[best_score_args]

        remaining_x_min = x_min[remaining_sorted_box_indices]
        remaining_y_min = y_min[remaining_sorted_box_indices]
        remaining_x_max = x_max[remaining_sorted_box_indices]
        remaining_y_max = y_max[remaining_sorted_box_indices]

        inner_x_min = np.maximum(remaining_x_min, best_x_min)
        inner_y_min = np.maximum(remaining_y_min, best_y_min)
        inner_x_max = np.minimum(remaining_x_max, best_x_max)
        inner_y_max = np.minimum(remaining_y_max, best_y_max)

        inner_box_widths = inner_x_max - inner_x_min
        inner_box_heights = inner_y_max - inner_y_min

        inner_box_widths = np.maximum(inner_box_widths, 0.0)
        inner_box_heights = np.maximum(inner_box_heights, 0.0)

        intersections = inner_box_widths * inner_box_heights
        remaining_box_areas = areas[remaining_sorted_box_indices]
        best_area = areas[best_score_args]
        unions = remaining_box_areas + best_area - intersections
        intersec_over_union = intersections / unions
        intersec_over_union_mask = intersec_over_union <= iou_thresh
        remaining_sorted_box_indices = remaining_sorted_box_indices[
            intersec_over_union_mask]

    return selected_indices.astype(int), num_selected_boxes


def nms_per_class(box_data, nms_thresh=.45, epsilon=0.01, top_k=200):
    """Applies non maximum suppression per class.
    This function takes all the detections from the detector which
    consists of boxes and their corresponding class scores to which it
    applies non maximum suppression for every class independently and
    then combines the result.

    # Arguments
        box_data: Array of shape `(num_nms_boxes, 4 + num_classes)`
            containing the box coordinates as well as the predicted
            scores of all the classes for all non suppressed boxes.
        nms_thresh: Float, Non-maximum suppression threshold.
        epsilon: Float, Filter scores with a lower confidence
            value before performing non-maximum supression.
        top_k: Int, Maximum number of boxes per class outputted by nms.

    # Returns
        Tuple: Containing an array non suppressed boxes of shape
            `(num_nms_boxes, 4 + num_classes)` and an array
            of corresponding class labels of shape `(num_nms_boxes, )`.
    """
    decoded_boxes = box_data[:, :4]
    class_predictions = box_data[:, 4:]
    num_classes = class_predictions.shape[1]
    nms_boxes = np.array([], dtype=float).reshape(0, box_data.shape[1])
    class_labels = np.array([], dtype=np.int)
    args = (decoded_boxes, class_predictions, epsilon, nms_thresh, top_k)
    for class_arg in range(num_classes):
        nms_boxes, class_labels = _nms_per_class(
            nms_boxes, class_labels, class_arg, *args)
    return nms_boxes, class_labels


def _nms_per_class(nms_boxes, class_labels, class_arg, decoded_boxes,
                   class_predictions, epsilon, nms_thresh, top_k):
    """Applies non maximum suppression for a given class.
    This function takes all the detections that belong only to the given
    single class and applies non maximum suppression for that class
    alone and returns the resulting non suppressed boxes.

    # Arguments
        nms_boxes: Array of shape `(num_boxes, 4 + num_classes)`.
        class_labels: Array of shape `(num_boxes, )`.
        class_arg: Int, class index.
        decoded_boxes: Array of shape `(num_prior_boxes, 4)`
            containing the box coordinates of all the
            non suppressed boxes.
        class_predictions: Array of shape
            `(num_nms_boxes, num_classes)` containing the predicted
            scores of all the classes for all the non suppressed boxes.
        epsilon: Float, Filter scores with a lower confidence
            value before performing non-maximum supression.
        nms_thresh: Float, Non-maximum suppression threshold.
        top_k: Int, Maximum number of boxes per class outputted by nms.

    # Returns
        Tuple: Containing an array non suppressed boxes per class of
            shape `(num_nms_boxes_per_class, 4 + num_classes) and an
            array corresponding class labels of shape
            `(num_nms_boxes_per_class, )`.
    """
    scores, mask = pre_filter_nms(class_arg, class_predictions, epsilon)

    if len(scores) != 0:
        boxes = decoded_boxes[mask]
        selected = apply_non_max_suppression(boxes, scores, nms_thresh, top_k)
        indices, count = selected
        selected_indices = indices[:count]
        selected_boxes = boxes[selected_indices]
        selected_classes = class_predictions[mask][selected_indices]
        selections = np.concatenate((selected_boxes, selected_classes), axis=1)
        nms_boxes = np.concatenate((nms_boxes, selections), axis=0)
        class_label = np.repeat(class_arg, count)
        class_labels = np.append(class_labels, class_label)
    return nms_boxes, class_labels


def pre_filter_nms(class_arg, class_predictions, epsilon):
    """Applies score filtering.
    This function takes all the predicted scores of a given class and
    filters out all the predictions less than the given `epsilon` value.

    # Arguments
        class_arg: Int, class index.
        class_predictions: Array of shape
            `(num_nms_boxes, num_classes)` containing the predicted
            scores of all the classes for all the non suppressed boxes.
        epsilon: Float, threshold value for score filtering.

    # Returns
        Tuple: Containing an array filtered scores of shape
            `(num_pre_filtered_boxes, )` and an array filter mask of
            shape `(num_prior_boxes, )`.
    """
    mask = class_predictions[:, class_arg] >= epsilon
    scores = class_predictions[:, class_arg][mask]
    return scores, mask


def merge_nms_box_with_class(box_data, class_labels):
    """Merges box coordinates with their corresponding class
    defined by `class_labels` which is decided by best box geometry
    by non maximum suppression (and not by the best scoring class)
    into a single output.
    This function retains only the predicted score of the class to
    which the box belongs to and sets the scores of all the remaining
    classes to zero, thereby combining box and class information in a
    single variable.

    # Arguments
        box_data: Array of shape `(num_nms_boxes, 4 + num_classes)`
            containing the box coordinates as well as the predicted
            scores of all the classes for all non suppressed boxes.
        class_labels: Array of shape `(num_nms_boxes, )` that contains
            the indices of the class whose score is to be retained.

    # Returns
        boxes: Array of shape `(num_nms_boxes, 4 + num_classes)`,
            containing coordinates of non supressed boxes along with
            scores of the class to which the box belongs. The scores of
            the other classes are zeros.
    """
    decoded_boxes = box_data[:, :4]
    class_predictions = box_data[:, 4:]
    retained_class_score = suppress_other_class_scores(
        class_predictions, class_labels)
    box_data = np.concatenate((decoded_boxes, retained_class_score), axis=1)
    return box_data


def suppress_other_class_scores(class_predictions, class_labels):
    """Retains the score of class in `class_labels` and
    sets other class scores to zero.

    # Arguments
        class_predictions: Array of shape
            `(num_nms_boxes, num_classes)` containing the predicted
            scores of all the classes for all the non suppressed boxes.
        class_labels: Array of shape `(num_nms_boxes, )` that contains
            the indices of the class whose score is to be retained.

    # Returns
        retained_class_score: Array of shape
            `(num_nms_boxes, num_classes)` that consists of score at
            only those location specified by 'class_labels' and zero
            at other class locations.

    # Note
        This approach retains the scores of that class in
        `class_predictions` defined by `class_labels` by generating
        a boolean mask `score_suppress_mask` with elements True at the
        locations where the score in `class_predictions` is to be
        retained and False wherever the class score is to be suppressed.
        This approach of retaining/suppressing scores does not make use
        of for loop, if-else condition and direct value assignment
        to arrays.
    """
    num_nms_boxes, num_classes = class_predictions.shape
    class_indices = np.arange(num_classes)
    class_indices = np.expand_dims(class_indices, axis=0)
    class_indices = np.repeat(class_indices, num_nms_boxes, axis=0)
    class_labels = np.expand_dims(class_labels, axis=1)
    class_labels = np.repeat(class_labels, num_classes, axis=1)
    """
    The difference of class_indices and class_labels contains zero
    at those locations of the result where the score is to be retained
    whose boolean value is False while others being True. This
    difference obtained as a boolean array gives a negative mask which
    when inverted gives the score_suppress_mask.
    """
    negative_mask = np.array(class_indices - class_labels, dtype=bool)
    score_suppress_mask = np.logical_not(negative_mask)
    retained_class_score = np.multiply(class_predictions, score_suppress_mask)
    return retained_class_score


def to_one_hot(class_indices, num_classes):
    """ Transform from class index to one-hot encoded vector.

    # Arguments
        class_indices: Numpy array. One dimensional array specifying
            the index argument of the class for each sample.
        num_classes: Integer. Total number of classes.

    # Returns
        Numpy array with shape `(num_samples, num_classes)`.
    """
    one_hot_vectors = np.zeros((len(class_indices), num_classes))
    for vector_arg, class_args in enumerate(class_indices):
        one_hot_vectors[vector_arg, class_args] = 1.0
    return one_hot_vectors


def make_box_square(box):
    """Makes box coordinates square with sides equal to the longest
        original side.

    # Arguments
        box: Numpy array with shape `(4)` with point corner coordinates.

    # Returns
        returns: List of box coordinates ints.
    """
    # TODO add ``calculate_center`` ``calculate_side_dimensions`` functions.
    x_min, y_min, x_max, y_max = box[:4]
    center_x = (x_max + x_min) / 2.0
    center_y = (y_max + y_min) / 2.0
    width = x_max - x_min
    height = y_max - y_min

    if height >= width:
        half_box = height / 2.0
        x_min = int(center_x - half_box)
        x_max = int(center_x + half_box)

    if width > height:
        half_box = width / 2.0
        y_min = int(center_y - half_box)
        y_max = int(center_y + half_box)

    return x_min, y_min, x_max, y_max


def offset(coordinates, offset_scales):
    """Apply offsets to box coordinates

    # Arguments
        coordinates: List of floats containing coordinates in point form.
        offset_scales: List of floats having x and y scales respectively.

    # Returns
        coordinates: List of floats containing coordinates in point form.
            i.e. [x_min, y_min, x_max, y_max].
    """
    x_min, y_min, x_max, y_max = coordinates
    x_offset_scale, y_offset_scale = offset_scales
    x_offset = (x_max - x_min) * x_offset_scale
    y_offset = (y_max - y_min) * y_offset_scale
    x_min = int(x_min - x_offset)
    y_max = int(y_max + x_offset)
    y_min = int(y_min - y_offset)
    x_max = int(x_max + y_offset)
    return (x_min, y_min, x_max, y_max)


def clip(coordinates, image_shape):
    """Clip box to valid image coordinates
    # Arguments
        coordinates: List of floats containing coordinates in point form
            i.e. [x_min, y_min, x_max, y_max].
        image_shape: List of two integers indicating height and width of image
            respectively.

    # Returns
        List of clipped coordinates.
    """
    height, width = image_shape[:2]
    x_min, y_min, x_max, y_max = coordinates
    if x_min < 0:
        x_min = 0
    if y_min < 0:
        y_min = 0
    if x_max > width:
        x_max = width
    if y_max > height:
        y_max = height
    return x_min, y_min, x_max, y_max


def denormalize_box(box, image_shape):
    """Scales corner box coordinates from normalized values to image dimensions

    # Arguments
        box: Numpy array containing corner box coordinates.
        image_shape: List of integers with (height, width).

    # Returns
        returns: box corner coordinates in image dimensions
    """
    x_min, y_min, x_max, y_max = box[:4]
    height, width = image_shape
    x_min = int(x_min * width)
    y_min = int(y_min * height)
    x_max = int(x_max * width)
    y_max = int(y_max * height)
    return (x_min, y_min, x_max, y_max)


def flip_left_right(boxes, width):
    """Flips box coordinates from left-to-right and vice-versa.
    # Arguments
        boxes: Numpy array of shape `[num_boxes, 4]`.
    # Returns
        Numpy array of shape `[num_boxes, 4]`.
    """
    flipped_boxes = boxes.copy()
    flipped_boxes[:, [0, 2]] = width - boxes[:, [2, 0]]
    return flipped_boxes


def to_image_coordinates(boxes, image):
    """Transforms normalized box coordinates into image coordinates.
    # Arguments
        image: Numpy array.
        boxes: Numpy array of shape `[num_boxes, N]` where N >= 4.
    # Returns
        Numpy array of shape `[num_boxes, N]`.
    """
    height, width = image.shape[:2]
    image_boxes = boxes.copy()
    image_boxes[:, 0] = boxes[:, 0] * width
    image_boxes[:, 2] = boxes[:, 2] * width
    image_boxes[:, 1] = boxes[:, 1] * height
    image_boxes[:, 3] = boxes[:, 3] * height
    return image_boxes


def to_normalized_coordinates(boxes, image):
    """Transforms coordinates in image dimensions to normalized coordinates.
    # Arguments
        image: Numpy array.
        boxes: Numpy array of shape `[num_boxes, N]` where N >= 4.
    # Returns
        Numpy array of shape `[num_boxes, N]`.
    """
    height, width = image.shape[:2]
    normalized_boxes = boxes.copy()
    normalized_boxes[:, 0] = boxes[:, 0] / width
    normalized_boxes[:, 2] = boxes[:, 2] / width
    normalized_boxes[:, 1] = boxes[:, 1] / height
    normalized_boxes[:, 3] = boxes[:, 3] / height
    return normalized_boxes


def extract_bounding_box_corners(points3D):
    """Extracts the (x_min, y_min, z_min) and the (x_max, y_max, z_max)
        coordinates from an array of  points3D
    # Arguments
        points3D: Array (num_points, 3)

    # Returns
        Left-down-bottom corner (x_min, y_min, z_min) and right-up-top
            (x_max, y_max, z_max) corner.
    """
    XYZ_min = np.min(points3D, axis=0)
    XYZ_max = np.max(points3D, axis=0)
    return XYZ_min, XYZ_max


def filter_boxes(boxes, conf_thresh):
    """Filters given boxes based on scores.

    # Arguments
        boxes: Array of shape `(num_nms_boxes, 4 + num_classes)`.
        conf_thresh: Float, Filter boxes with a confidence value
            lower than this.

    # Returns
        confident_boxes: Array of shape
            `(num_filtered_boxes, 4 + num_classes)`.
    """
    class_predictions = boxes[:, 4:]
    class_scores = np.max(class_predictions, axis=1)
    confidence_mask = class_scores >= conf_thresh
    confident_boxes = boxes[confidence_mask]
    return confident_boxes


def scale_box(predictions, image_scales):
    """
    # Arguments
        predictions: Array of shape `(num_boxes, num_classes+N)`
            model predictions.
        image_scales: Array of shape `()`, scale value of boxes.

    # Returns
        predictions: Array of shape `(num_boxes, num_classes+N)`
            model predictions.
    """
    boxes = predictions[:, :4]
    scales = image_scales[np.newaxis][np.newaxis]
    boxes = boxes * scales
    predictions = np.concatenate([boxes, predictions[:, 4:]], 1)
    return predictions


def change_box_coordinates(outputs):
    """Converts box coordinates format from (y_min, x_min, y_max, x_max)
    to (x_min, y_min, x_max, y_max).

    # Arguments
        outputs: Tensor, model output.

    # Returns
        outputs: Array, Processed outputs by merging the features
            at all levels. Each row corresponds to box coordinate
            offsets and sigmoid of the class logits.
    """
    outputs = outputs[0]
    boxes, classes = outputs[:, :4], outputs[:, 4:]
    s1, s2, s3, s4 = np.hsplit(boxes, 4)
    boxes = np.concatenate([s2, s1, s4, s3], axis=1)
    boxes = boxes[np.newaxis]
    classes = classes[np.newaxis]
    outputs = np.concatenate([boxes, classes], axis=2)
    return outputs



import numpy as np

from tensorflow.keras.utils import get_file

from .renderer import RenderTwoViews
from .image import PreprocessImageHigherHRNet
from .heatmaps import GetHeatmapsAndTags

from .. import processors as pr
from ..abstract import SequentialProcessor, Processor
from ..models import KeypointNet2D, HigherHRNet, DetNet, SimpleBaseline
from .angles import IKNetHandJointAngles

from ..backend.image import get_affine_transform, lincolor
from ..backend.keypoints import flip_keypoints_left_right, uv_to_vu
from ..datasets import JOINT_CONFIG, FLIP_CONFIG

from ..datasets.human36m import data_mean2D, data_stdev2D, args_to_mean
from ..datasets.human36m import data_mean3D, data_stdev3D, dim_to_use3D
from ..datasets.human36m import h36m_to_coco_joints2D


class KeypointNetSharedAugmentation(SequentialProcessor):
    """Wraps ``RenderTwoViews`` as a sequential processor for using it directly
        with a ``paz.GeneratingSequence``.

    # Arguments
        renderer: ``RenderTwoViews`` processor.
        size: Image size.
    """
    def __init__(self, renderer, size):
        super(KeypointNetSharedAugmentation, self).__init__()
        self.renderer = renderer
        self.size = size
        self.add(RenderTwoViews(self.renderer))
        self.add(pr.SequenceWrapper(
            {0: {'image_A': [size, size, 3]},
             1: {'image_B': [size, size, 3]}},
            {2: {'matrices': [4, 4 * 4]},
             3: {'alpha_channels': [size, size, 2]}}))


class KeypointNetInference(Processor):
    """Performs inference from a ``KeypointNetShared`` model.

    # Arguments
        model: Keras model for predicting keypoints.
        num_keypoints: Int or None. If None ``num_keypoints`` is
            tried to be inferred from ``model.output_shape``
        radius: Int. used for drawing the predicted keypoints.
    """
    def __init__(self, model, num_keypoints=None, radius=5):
        super(KeypointNetInference, self).__init__()
        self.num_keypoints, self.radius = num_keypoints, radius
        if self.num_keypoints is None:
            self.num_keypoints = model.output_shape[1]

        preprocessing = SequentialProcessor()
        preprocessing.add(pr.NormalizeImage())
        preprocessing.add(pr.ExpandDims(axis=0))
        self.predict_keypoints = SequentialProcessor()
        self.predict_keypoints.add(pr.Predict(model, preprocessing))
        self.predict_keypoints.add(pr.SelectElement(0))
        self.predict_keypoints.add(pr.Squeeze(axis=0))
        self.postprocess_keypoints = SequentialProcessor()
        self.postprocess_keypoints.add(pr.DenormalizeKeypoints())
        self.postprocess_keypoints.add(pr.RemoveKeypointsDepth())
        self.draw = pr.DrawKeypoints2D(self.num_keypoints, self.radius, False)
        self.wrap = pr.WrapOutput(['image', 'keypoints'])

    def call(self, image):
        keypoints = self.predict_keypoints(image)
        keypoints = self.postprocess_keypoints(keypoints, image)
        image = self.draw(image, keypoints)
        return self.wrap(image, keypoints)


class EstimateKeypoints2D(Processor):
    """Basic 2D keypoint prediction pipeline.

    # Arguments
        model: Keras model for predicting keypoints.
        num_keypoints: Int or None. If None ``num_keypoints`` is
            tried to be inferred from ``model.output_shape``
        draw: Boolean indicating if inferences should be drawn.
        radius: Int. used for drawing the predicted keypoints.
    """
    def __init__(self, model, num_keypoints, draw=True, radius=3,
                 color=pr.RGB2BGR):
        self.model = model
        self.num_keypoints = num_keypoints
        self.draw, self.radius, self.color = draw, radius, color
        self.preprocess = SequentialProcessor()
        self.preprocess.add(pr.ResizeImage(self.model.input_shape[1:3]))
        self.preprocess.add(pr.ConvertColorSpace(self.color))
        self.preprocess.add(pr.NormalizeImage())
        self.preprocess.add(pr.ExpandDims(0))
        self.preprocess.add(pr.ExpandDims(-1))
        self.predict = pr.Predict(model, self.preprocess, pr.Squeeze(0))
        self.denormalize = pr.DenormalizeKeypoints()
        self.draw = pr.DrawKeypoints2D(self.num_keypoints, self.radius, False)
        self.wrap = pr.WrapOutput(['image', 'keypoints'])

    def call(self, image):
        keypoints = self.predict(image)
        keypoints = self.denormalize(keypoints, image)
        if self.draw:
            image = self.draw(image, keypoints)
        return self.wrap(image, keypoints)


class FaceKeypointNet2D32(EstimateKeypoints2D):
    """KeypointNet2D model trained with Kaggle Facial Detection challenge.

    # Arguments
        draw: Boolean indicating if inferences should be drawn.
        radius: Int. used for drawing the predicted keypoints.

    # Example
        ``` python
        from paz.pipelines import FaceKeypointNet2D32

        estimate_keypoints= FaceKeypointNet2D32()

        # apply directly to an image (numpy-array)
        inference = estimate_keypoints(image)
        ```
    # Returns
        A function that takes an RGB image and outputs the predictions
        as a dictionary with ``keys``: ``image`` and ``keypoints``.
        The corresponding values of these keys contain the image with the drawn
        inferences and a numpy array representing the keypoints.
    """
    def __init__(self, draw=True, radius=3):
        model = KeypointNet2D((96, 96, 1), 15, 32, 0.1)
        self.weights_URL = ('https://github.com/oarriaga/altamira-data/'
                            'releases/download/v0.7/')
        weights_path = self.get_weights_path(model)
        model.load_weights(weights_path)
        super(FaceKeypointNet2D32, self).__init__(
            model, 15, draw, radius, pr.RGB2GRAY)

    def get_weights_path(self, model):
        model_name = '_'.join(['FaceKP', model.name, '32', '15'])
        model_name = '%s_weights.hdf5' % model_name
        URL = self.weights_URL + model_name
        return get_file(model_name, URL, cache_subdir='paz/models')


class GetKeypoints(Processor):
    """Extract out the top k keypoints heatmaps and group the keypoints with
       their respective tags value. Adjust and refine the keypoint locations
       by removing the margins.
    # Arguments
        max_num_instance: Int. Maximum number of instances to be detected.
        keypoint_order: List of length 17 (number of keypoints).
        heatmaps: Numpy array of shape (1, num_keypoints, H, W)
        Tags: Numpy array of shape (1, num_keypoints, H, W, 2)

    # Returns
        grouped_keypoints: numpy array. keypoints grouped by tag
        scores: int: score for the keypoint
    """
    def __init__(self, max_num_instance, keypoint_order, detection_thresh=0.2,
                 tag_thresh=1):
        super(GetKeypoints, self).__init__()
        self.group_keypoints = pr.SequentialProcessor(
            [pr.TopKDetections(max_num_instance), pr.GroupKeypointsByTag(
                keypoint_order, tag_thresh, detection_thresh)])
        self.adjust_keypoints = pr.AdjustKeypointsLocations()
        self.get_scores = pr.GetScores()
        self.refine_keypoints = pr.RefineKeypointsLocations()

    def call(self, heatmaps, tags, adjust=True, refine=True):
        grouped_keypoints = self.group_keypoints(heatmaps, tags)
        if adjust:
            grouped_keypoints = self.adjust_keypoints(
                heatmaps, grouped_keypoints)[0]
        scores = self.get_scores(grouped_keypoints)
        if refine:
            grouped_keypoints = self.refine_keypoints(
                heatmaps[0], tags[0], grouped_keypoints)
        return grouped_keypoints, scores


class TransformKeypoints(Processor):
    """Transform the keypoint coordinates.
    # Arguments
        grouped_keypoints: Numpy array. keypoints grouped by tag
        center: Tuple. center of the imput image
        scale: Float. scaled imput image dimension
        shape: Tuple/List

    # Returns
        transformed_keypoints: keypoint location with respect to the
                               input image
    """
    def __init__(self, inverse=False):
        super(TransformKeypoints, self).__init__()
        self.inverse = inverse
        self.get_source_destination_point = pr.GetSourceDestinationPoints(
            scaling_factor=200)
        self.transform_keypoints = pr.TransformKeypoints()

    def call(self, grouped_keypoints, center, scale, shape):
        source_point, destination_point = self.get_source_destination_point(
            center, scale, shape)
        if self.inverse:
            source_point, destination_point = destination_point, source_point
        transform = get_affine_transform(source_point, destination_point)
        transformed_keypoints = self.transform_keypoints(grouped_keypoints,
                                                         transform)
        return transformed_keypoints


class HigherHRNetHumanPose2D(Processor):
    """Estimate human pose 2D keypoints and draw a skeleton.

    # Arguments
        model: Weights trained on HigherHRNet model.
        keypoint_order: List of length 17 (number of keypoints).
            where the keypoints are listed order wise.
        flipped_keypoint_order: List of length 17 (number of keypoints).
            Flipped list of keypoint order.
        dataset: String. Name of the dataset used for training the model.
        data_with_center: Boolean. True is the model is trained using the
            center.

    # Returns
        dictonary with the following keys:
            image: contains the image with skeleton drawn on it.
            keypoints: location of keypoints
            score: score of detection
    """
    def __init__(self, dataset='COCO', data_with_center=False,
                 max_num_people=30, with_flip=True, draw=True):
        super(HigherHRNetHumanPose2D, self).__init__()
        keypoint_order = JOINT_CONFIG[dataset]
        flipped_keypoint_order = FLIP_CONFIG[dataset]
        self.with_flip = with_flip
        self.draw = draw
        self.model = HigherHRNet(weights=dataset)
        self.transform_image = PreprocessImageHigherHRNet()
        self.get_heatmaps_and_tags = pr.SequentialProcessor(
            [GetHeatmapsAndTags(self.model, flipped_keypoint_order,
             with_flip, data_with_center), pr.AggregateResults(with_flip)])
        self.get_keypoints = GetKeypoints(max_num_people, keypoint_order)
        self.transform_keypoints = TransformKeypoints(inverse=True)
        self.draw_skeleton = pr.DrawHumanSkeleton(dataset, check_scores=True)
        self.extract_keypoints_locations = pr.ExtractKeypointsLocations()
        self.wrap = pr.WrapOutput(['image', 'keypoints', 'scores'])

    def call(self, image):
        resized_image, center, scale = self.transform_image(image)
        heatmaps, tags = self.get_heatmaps_and_tags(resized_image)
        keypoints, scores = self.get_keypoints(heatmaps, tags)
        shape = [heatmaps.shape[3], heatmaps.shape[2]]
        keypoints = self.transform_keypoints(keypoints, center, scale, shape)
        if self.draw:
            image = self.draw_skeleton(image, keypoints)
        keypoints = self.extract_keypoints_locations(keypoints)
        return self.wrap(image, keypoints, scores)


class DetNetHandKeypoints(pr.Processor):
    """Estimate 2D and 3D keypoints from minimal hand and draw a skeleton.

    # Arguments
        shape: List/tuple. Input image shape for DetNet model.
        draw: Boolean. Draw hand skeleton if true.
        right_hand: Boolean. If 'True', detect keypoints for right hand, else
                    detect keypoints for left hand.
        input_image: Array

    # Returns
        image: contains the image with skeleton drawn on it.
        keypoints2D: Array [num_joints, 2]. 2D location of keypoints.
        keypoints3D: Array [num_joints, 3]. 3D location of keypoints.
    """
    def __init__(self, shape=(128, 128), draw=True, right_hand=False):
        super(DetNetHandKeypoints).__init__()
        self.draw = draw
        self.right_hand = right_hand
        self.preprocess = pr.SequentialProcessor()
        self.preprocess.add(pr.ResizeImage(shape))
        self.preprocess.add(pr.ExpandDims(axis=0))
        if self.right_hand:
            self.preprocess.add(pr.FlipLeftRightImage())
        self.predict = pr.Predict(model=DetNet(), preprocess=self.preprocess)
        self.scale_keypoints = pr.ScaleKeypoints(scale=4, shape=shape)
        self.draw_skeleton = pr.DrawHandSkeleton()
        self.wrap = pr.WrapOutput(['image', 'keypoints3D', 'keypoints2D'])

    def call(self, image):
        keypoints3D, keypoints2D = self.predict(image)
        keypoints3D = keypoints3D.numpy()
        keypoints2D = keypoints2D.numpy()
        if self.right_hand:
            keypoints2D = flip_keypoints_left_right(keypoints2D)
        keypoints2D = uv_to_vu(keypoints2D)
        keypoints2D = self.scale_keypoints(keypoints2D, image)
        if self.draw:
            image = self.draw_skeleton(image, keypoints2D)
        return self.wrap(image, keypoints3D, keypoints2D)


class MinimalHandPoseEstimation(pr.Processor):
    """Estimate 2D and 3D keypoints from minimal hand and draw a skeleton.
       Estimate absolute and relative joint angle for the minimal hand joints
       using the 3D keypoint locations.

    # Arguments
        draw: Boolean. Draw hand skeleton if true.
        right_hand: Boolean. If 'True', detect keypoints for right hand, else
                    detect keypoints for left hand.

    # Returns
        image: contains the image with skeleton drawn on it.
        keypoints2D: Array [num_joints, 2]. 2D location of keypoints.
        keypoints3D: Array [num_joints, 3]. 3D location of keypoints.
        absolute_angles: Array [num_joints, 4]. quaternion repesentation
        relative_angles: Array [num_joints, 3]. axis-angle repesentation
    """
    def __init__(self, draw=True, right_hand=False):
        super(MinimalHandPoseEstimation, self).__init__()
        self.keypoints_estimator = DetNetHandKeypoints(draw=draw,
                                                       right_hand=right_hand)
        self.angle_estimator = IKNetHandJointAngles(right_hand=right_hand)
        self.wrap = pr.WrapOutput(['image', 'keypoints3D', 'keypoints2D',
                                   'absolute_angles', 'relative_angles'])

    def call(self, image):
        keypoints = self.keypoints_estimator(image)
        angles = self.angle_estimator(keypoints['keypoints3D'])
        return self.wrap(keypoints['image'], keypoints['keypoints3D'],
                         keypoints['keypoints2D'], angles['absolute_angles'],
                         angles['relative_angles'])


class DetectMinimalHand(pr.Processor):
    def __init__(self, detect, estimate_keypoints, offsets=[0, 0], radius=3):
        """Minimal hand detection and keypoint estimator pipeline.

        # Arguments
            detect: Function for detecting objects. The output should be a
                dictionary with key ``Boxes2D`` containing a list
                of ``Boxes2D`` messages.
            estimate_keypoints: Function for estimating keypoints. The output
                should be a dictionary with key ``keypoints`` containing
                a numpy array of keypoints.
            offsets: List of two elements. Each element must be between [0, 1].
            radius: Int indicating the radius of the keypoints to be drawn.
        """
        super(DetectMinimalHand, self).__init__()
        self.class_names = ['OPEN', 'CLOSE']
        self.colors = lincolor(len(self.class_names))
        self.detect = detect
        self.estimate_keypoints = estimate_keypoints
        self.classify_hand_closure = pr.SequentialProcessor(
            [pr.IsHandOpen(), pr.BooleanToTextMessage('OPEN', 'CLOSE')])
        self.square = pr.SequentialProcessor()
        self.square.add(pr.SquareBoxes2D())
        self.square.add(pr.OffsetBoxes2D(offsets))
        self.clip = pr.ClipBoxes2D()
        self.crop = pr.CropBoxes2D()
        self.change_coordinates = pr.ChangeKeypointsCoordinateSystem()
        self.draw = pr.DrawHandSkeleton(keypoint_radius=radius)
        self.draw_boxes = pr.DrawBoxes2D(self.class_names, self.colors,
                                         with_score=False)
        self.wrap = pr.WrapOutput(
            ['image', 'boxes2D', 'keypoints2D', 'keypoints3D'])

    def call(self, image):
        boxes2D = self.detect(image.copy())['boxes2D']
        boxes2D = self.square(boxes2D)
        boxes2D = self.clip(image, boxes2D)
        cropped_images = self.crop(image, boxes2D)
        keypoints2D = []
        keypoints3D = []
        for cropped_image, box2D in zip(cropped_images, boxes2D):
            inference = self.estimate_keypoints(cropped_image)
            keypoints = self.change_coordinates(
                inference['keypoints2D'], box2D)
            hand_closure_status = self.classify_hand_closure(
                inference['relative_angles'])
            box2D.class_name = hand_closure_status
            keypoints2D.append(keypoints)
            keypoints3D.append(inference['keypoints3D'])
            image = self.draw(image, keypoints)
        image = self.draw_boxes(image, boxes2D)
        return self.wrap(image, boxes2D, keypoints2D, keypoints3D)


class EstimateHumanPose3D(Processor):
    """ Estimate human pose 3D from 2D human pose.

    # Arguments
    input_shape: tuple
    num_keypoints: Int. Number of keypoints.

    # Return
        keypoints3D: human pose 3D
    """
    def __init__(self, input_shape=(32,), num_keypoints=16):
        super(EstimateHumanPose3D, self).__init__()
        self.model = SimpleBaseline(input_shape, num_keypoints)
        self.preprocessing = SequentialProcessor(
            [pr.MergeKeypoints2D(args_to_mean),
             pr.FilterKeypoints2D(args_to_mean, h36m_to_coco_joints2D),
             pr.StandardizeKeypoints2D(data_mean2D, data_stdev2D)])
        self.postprocess = pr.DestandardizeKeypoints2D(
            data_mean3D, data_stdev3D, dim_to_use3D)
        self.predict = pr.Predict(self.model, self.preprocessing,
                                  self.postprocess)

    def call(self, keypoints2D):
        keypoints3D = self.predict(keypoints2D)
        return keypoints3D


class EstimateHumanPose(pr.Processor):
    """ Estimates 2D and 3D keypoints of human from an image

    # Arguments
        estimate_keypoints_3D: 3D simple baseline model
        args_to_mean: keypoints indices
        h36m_to_coco_joints2D: h36m joints indices

    # Returns
        keypoints2D, keypoints3D
    """
    def __init__(self, filter=True):
        super(EstimateHumanPose, self).__init__()
        self.filter = filter
        self.estimate_keypoints_2D = HigherHRNetHumanPose2D()
        self.filter_keypoints2D = SequentialProcessor(
            [pr.MergeKeypoints2D(args_to_mean),
             pr.FilterKeypoints2D(args_to_mean, h36m_to_coco_joints2D)])
        self.estimate_keypoints_3D = EstimateHumanPose3D()
        self.wrap = pr.WrapOutput(['keypoints2D', 'keypoints3D'])

    def call(self, image):
        inferences2D = self.estimate_keypoints_2D(image)
        keypoints2D = np.array(inferences2D['keypoints'])
        if self.filter:
            filter_keypoints2D = self.filter_keypoints2D(keypoints2D)
        keypoints3D = self.estimate_keypoints_3D(keypoints2D)
        keypoints3D = np.reshape(keypoints3D, (-1, 32, 3))
        return self.wrap(filter_keypoints2D, keypoints3D)



from tensorflow.keras import backend as K
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Input
from tensorflow.keras.layers import UpSampling2D
from tensorflow.keras.layers import Concatenate
from tensorflow.keras.layers import Permute
from tensorflow.keras.layers import Reshape
from tensorflow.keras.layers import Add
from tensorflow.keras.models import Model
from ..layers import ExpectedValue2D


def dense_block(x, blocks, growth_rate):
    for block_arg in range(blocks):
        x1 = Conv2D(4 * growth_rate, 1, use_bias=False)(x)
        x1 = BatchNormalization(epsilon=1.001e-5)(x)
        x1 = Activation('relu')(x1)
        x1 = Conv2D(growth_rate, 3, padding='same', use_bias=False)(x1)
        x1 = BatchNormalization(epsilon=1.001e-5)(x1)
        x1 = Activation('relu')(x1)
        x = Concatenate(axis=-1)([x, x1])
    return x


def residual_block(x, num_kernels, strides=1):
    residual = x
    x = Conv2D(num_kernels, 3, padding='same', use_bias=False)(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    x = Conv2D(num_kernels, 3, padding='same', use_bias=False)(x)
    x = BatchNormalization()(x)
    x = Add()([x, residual])
    x = Activation('relu')(x)
    return x


def transition_block(x, alpha):
    filters = int(K.int_shape(x)[-1] * alpha)
    x = Conv2D(filters, 1, strides=2, use_bias=False)(x)
    x = BatchNormalization(epsilon=1.001e-5)(x)
    x = Activation('relu')(x)
    return x


def stem(x, filters):
    x = Conv2D(filters, 3, padding='same', strides=(2, 2), use_bias=False)(x)
    x = BatchNormalization(epsilon=1.1e-5)(x)
    x = Activation('relu')(x)
    x = Conv2D(filters, 3, padding='same', strides=(2, 2), use_bias=False)(x)
    x = BatchNormalization(epsilon=1.1e-5)(x)
    x = Activation('relu')(x)
    return x


def fuse(tensors, base_kernels=32):
    all_tensors = []
    for x_tensor_arg, x in enumerate(tensors):
        x_to_y_tensors = []
        for y_tensor_arg in range(len(tensors)):
            # step: how much the feature map is upsampled or downsampled
            steps = x_tensor_arg - y_tensor_arg

            if steps == 0:
                num_kernels = K.int_shape(x)[-1]
                y = Conv2D(num_kernels, 3, padding='same',
                           strides=1, use_bias=False)(x)
                y = BatchNormalization(epsilon=1.1e-5)(y)
                y = Activation('relu')(y)

            if steps < 0:
                y = x
                for step in range(abs(steps)):
                    num_kernels = int(K.int_shape(x)[-1] * (step + 1))
                    y = Conv2D(num_kernels, 3, strides=2,
                               padding='same', use_bias=False)(y)
                    y = BatchNormalization(epsilon=1.1e-5)(y)
                    y = Activation('relu')(y)

            if steps > 0:
                num_kernels = int(K.int_shape(x)[-1] / steps)
                y = Conv2D(num_kernels, 1, use_bias=False)(x)
                y = BatchNormalization(epsilon=1.1e-5)(y)
                y = Activation('relu')(y)
                y = UpSampling2D(size=(2**steps, 2**steps))(y)

            x_to_y_tensors.append(y)
        all_tensors.append(x_to_y_tensors)

    output_tensors = []
    for reciever_arg in range(len(tensors)):
        same_resolution_tensors = []
        for giver_arg in range(len(tensors)):
            tensor = all_tensors[giver_arg][reciever_arg]
            same_resolution_tensors.append(tensor)
        x = Concatenate()(same_resolution_tensors)
        num_kernels = base_kernels * (2 ** (reciever_arg))
        x = Conv2D(num_kernels, 1, use_bias=False)(x)
        x = BatchNormalization(epsilon=1.1e-5)(x)
        x = Activation('relu')(x)
        output_tensors.append(x)
    return output_tensors


def bottleneck(x, filters=64, expansion=4):
    residual = x
    x = Conv2D(filters, 1, use_bias=False)(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    x = Conv2D(filters, 3, padding='same', use_bias=False)(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    x = Conv2D(filters * expansion, 1, use_bias=False)(x)
    x = BatchNormalization()(x)
    x = Add()([x, residual])
    x = Activation('relu')(x)
    return x


def HRNetDense(input_shape=(128, 128, 3), num_keypoints=20, growth_rate=4):
    # stem
    inputs = Input(shape=input_shape)
    x1 = stem(inputs, 64)
    x1 = Conv2D(64 * 4, 1, padding='same', use_bias=False)(x1)
    x1 = BatchNormalization()(x1)
    for block in range(4):
        x1 = bottleneck(x1)

    # stage I
    x1 = Conv2D(32, 3, padding='same', use_bias=False)(x1)
    x1 = BatchNormalization()(x1)
    x1 = Activation('relu')(x1)
    x2 = transition_block(x1, 2)
    print('stage 1', x1.shape, x2.shape)

    # stage II
    x1 = dense_block(x1, 4, growth_rate)
    x2 = dense_block(x2, 4, growth_rate)
    x1, x2 = fuse([x1, x2])
    x3 = transition_block(x2, 0.5)
    print('stage 2', x1.shape, x2.shape, x3.shape)

    # stage III
    x1 = dense_block(x1, 4, growth_rate)
    x2 = dense_block(x2, 4, growth_rate)
    x3 = dense_block(x3, 4, growth_rate)
    x1, x2, x3 = fuse([x1, x2, x3])
    x4 = transition_block(x3, 0.5)
    print('stage 3', x1.shape, x2.shape, x3.shape, x4.shape)

    # stage IV
    x1 = dense_block(x1, 3, growth_rate)
    x2 = dense_block(x2, 3, growth_rate)
    x3 = dense_block(x3, 3, growth_rate)
    x4 = dense_block(x4, 3, growth_rate)
    x1, x2, x3, x4 = fuse([x1, x2, x3, x4])
    print('stage 4', x1.shape, x2.shape, x3.shape, x4.shape)

    x2 = UpSampling2D(size=(2, 2))(x2)
    x3 = UpSampling2D(size=(4, 4))(x3)
    x4 = UpSampling2D(size=(8, 8))(x4)
    x = Concatenate()([x1, x2, x3, x4])

    # head
    x = Conv2D(480, 1)(x)
    x = BatchNormalization(epsilon=1.001e-5)(x)
    x = Activation('relu')(x)
    x = Conv2D(num_keypoints, 1)(x)

    # extra
    x = BatchNormalization(epsilon=1.001e-5)(x)
    x = Activation('relu')(x)
    x = UpSampling2D(size=(4, 4), interpolation='bilinear')(x)
    x = Permute([3, 1, 2])(x)
    x = Reshape([num_keypoints, input_shape[0] * input_shape[1]])(x)
    x = Activation('softmax')(x)
    x = Reshape([num_keypoints, input_shape[0], input_shape[1]])(x)
    outputs = ExpectedValue2D(name='expected_uv')(x)
    model = Model(inputs, outputs, name='hrnet-dense')
    return model


def HRNetResidual(input_shape=(128, 128, 3), num_keypoints=20):
    """Instantiates HRNET Residual model

    # Arguments
        input_shape: List of three elements e.g. ''(H, W, 3)''
        num_keypoints: Int.

    # Returns
        Tensorflow-Keras model.

    # References
       -[High-Resolution Representations for Labeling Pixels
            and Regions](https://arxiv.org/pdf/1904.04514.pdf)
    """

    # stem
    inputs = Input(shape=input_shape, name='image')
    x1 = stem(inputs, 64)
    x1 = Conv2D(64 * 4, 1, padding='same', use_bias=False)(x1)
    x1 = BatchNormalization()(x1)
    for block in range(4):
        x1 = bottleneck(x1)

    # stage I
    x1 = Conv2D(32, 3, padding='same', use_bias=False)(x1)
    x1 = BatchNormalization()(x1)
    x1 = Activation('relu')(x1)
    x2 = transition_block(x1, 2)

    # stage II
    for block in range(4):
        x1 = residual_block(x1, 32)
        x2 = residual_block(x2, 64)
    x1, x2 = fuse([x1, x2])
    x3 = transition_block(x2, 2)

    # stage III
    for module in range(4):
        for block in range(4):
            x1 = residual_block(x1, 32)
            x2 = residual_block(x2, 64)
            x3 = residual_block(x3, 128)
        x1, x2, x3 = fuse([x1, x2, x3])
    x4 = transition_block(x3, 2)

    # stage IV
    for module in range(3):
        for block in range(4):
            x1 = residual_block(x1, 32)
            x2 = residual_block(x2, 64)
            x3 = residual_block(x3, 128)
            x4 = residual_block(x4, 256)
        x1, x2, x3, x4 = fuse([x1, x2, x3, x4])

    # head
    x2 = UpSampling2D(size=(2, 2))(x2)
    x3 = UpSampling2D(size=(4, 4))(x3)
    x4 = UpSampling2D(size=(8, 8))(x4)
    x = Concatenate()([x1, x2, x3, x4])
    x = Conv2D(480, 1)(x)
    x = BatchNormalization(epsilon=1.001e-5)(x)
    x = Activation('relu')(x)
    x = Conv2D(num_keypoints, 1)(x)

    # extra
    x = BatchNormalization(epsilon=1.001e-5)(x)
    x = Activation('relu')(x)
    x = UpSampling2D(size=(4, 4), interpolation='bilinear')(x)
    x = Permute([3, 1, 2])(x)
    x = Reshape([num_keypoints, input_shape[0] * input_shape[1]])(x)
    x = Activation('softmax')(x)
    x = Reshape([num_keypoints, input_shape[0], input_shape[1]])(x)
    outputs = ExpectedValue2D(name='keypoints')(x)
    model = Model(inputs, outputs, name='hrnet-residual')
    return model



import tensorflow as tf
from tensorflow.keras import Model
from tensorflow.keras.layers import (Input, Conv2D, BatchNormalization, Layer,
                                     ReLU, MaxPool2D, Flatten, Softmax)

from ...utils.documentation import docstring


def conv_block(x):
    """Basic convolution block used for prototypical networks.

    # Arguments
        x: Tensor.

    # Returns
        Tensor
    """
    x = Conv2D(filters=64, kernel_size=3, padding='same')(x)
    x = BatchNormalization()(x)
    x = ReLU()(x)
    x = MaxPool2D((2, 2))(x)
    return x


def ProtoEmbedding(image_shape, num_blocks):
    """Embedding convolutional network used for proto-typical networks

    # Arguments:
        image_shape: List with image shape `(H, W, channels)`.
        num_blocks: Ints. Number of convolution blocks.

    # Returns:
        Keras model.

    # References:
        [prototypical networks](https://arxiv.org/abs/1703.05175)
    """
    x = inputs = Input(image_shape)
    for _ in range(num_blocks):
        x = conv_block(x)
    z = Flatten()(x)
    return Model(inputs, z, name='EMBEDDING')


class FullReshape(Layer):
    """Reshapes all tensor dimensions including the batch dimension.
    """
    def __init__(self, shape, **kwargs):
        super(FullReshape, self).__init__(**kwargs)
        self.shape = shape

    def call(self, x):
        return tf.reshape(x, self.shape)


class ComputePrototypes(Layer):
    def __init__(self, axis=1, **kwargs):
        super(ComputePrototypes, self).__init__(**kwargs)
        self.axis = axis

    def call(self, z_support):
        class_prototypes = tf.reduce_mean(z_support, axis=self.axis)
        return class_prototypes


def compute_pairwise_distances(x, y):
    """Compute euclidean distance for each vector x with each vector y

    # Arguments:
        x: Tensor with shape `(n, vector_dim)`
        y: Tensor with shape `(m, vector_dim)`

    # Returns:
        Tensor with shape `(n, m)` where each value pair n, m corresponds to
        the distance between the vector `n` of `x` with the vector `m` of `y`
    """
    n = x.shape[0]
    m = y.shape[0]
    x = tf.tile(tf.expand_dims(x, 1), [1, m, 1])
    y = tf.tile(tf.expand_dims(y, 0), [n, 1, 1])
    return tf.reduce_mean(tf.math.pow(x - y, 2), 2)


@docstring(compute_pairwise_distances)
class ComputePairwiseDistances(Layer):
    def __init__(self, **kwargs):
        super(ComputePairwiseDistances, self).__init__(**kwargs)

    def call(self, z_queries, class_prototypes):
        return compute_pairwise_distances(z_queries, class_prototypes)


def ProtoNet(embed, num_classes, num_support, num_queries, image_shape):
    """Prototypical networks used for few-shot classification
    # Arguments:
        embed: Keras network for embedding images into metric space.
        num_classes: Number of `ways` for few-shot classification.
        num_support: Number of `shots` used for meta learning.
        num_queries: Number of test images to query.
        image_shape: List with image shape `(H, W, channels)`.

    # Returns:
        Keras model.

    # References:
        [prototypical networks](https://arxiv.org/abs/1703.05175)
    """
    support = Input((num_support, *image_shape), num_classes, name='support')
    queries = Input((num_queries, *image_shape), num_classes, name='queries')
    z_support = FullReshape((num_classes * num_support, *image_shape))(support)
    z_queries = FullReshape((num_classes * num_queries, *image_shape))(queries)
    z_support = embed(z_support)
    z_queries = embed(z_queries)
    z_dim = embed.output_shape[-1]
    z_support = FullReshape((num_classes, num_support, z_dim))(z_support)
    z_queries = FullReshape((num_classes * num_queries, z_dim))(z_queries)
    class_prototypes = ComputePrototypes(axis=1)(z_support)
    distances = ComputePairwiseDistances()(z_queries, class_prototypes)
    outputs = Softmax()(-distances)
    return Model(inputs=[support, queries], outputs=outputs, name='PROTONET')



import numpy as np
import cv2
import os

RGB2BGR = cv2.COLOR_RGB2BGR
BGR2RGB = cv2.COLOR_BGR2RGB
BGRA2RGBA = cv2.COLOR_BGRA2RGBA
RGB2GRAY = cv2.COLOR_RGB2GRAY
RGB2HSV = cv2.COLOR_RGB2HSV
HSV2RGB = cv2.COLOR_HSV2RGB
_CHANNELS_TO_FLAG = {1: cv2.IMREAD_GRAYSCALE,
                     3: cv2.IMREAD_COLOR,
                     4: cv2.IMREAD_UNCHANGED}
CUBIC = cv2.INTER_CUBIC
BILINEAR = cv2.INTER_LINEAR


def resize_image(image, size, method=BILINEAR):
    """Resize image.

    # Arguments
        image: Numpy array.
        size: List of two ints.
        method: Flag indicating interpolation method i.e.
            paz.backend.image.CUBIC

    # Returns
        Numpy array.
    """
    if (type(image) != np.ndarray):
        raise ValueError(
            'Recieved Image is not of type numpy array', type(image))
    else:
        return cv2.resize(image, size, interpolation=method)


def convert_color_space(image, flag):
    """Convert image to a different color space.

    # Arguments
        image: Numpy array.
        flag: PAZ or openCV flag. e.g. paz.backend.image.RGB2BGR.

    # Returns
        Numpy array.
    """
    return cv2.cvtColor(image, flag)


def load_image(filepath, num_channels=3):
    """Load image from a ''filepath''.

    # Arguments
        filepath: String indicating full path to the image.
        num_channels: Int.

    # Returns
        Numpy array.
    """
    if num_channels not in [1, 3, 4]:
        raise ValueError('Invalid number of channels')

    image = cv2.imread(filepath, _CHANNELS_TO_FLAG[num_channels])
    if num_channels == 3:
        image = convert_color_space(image, BGR2RGB)
    elif num_channels == 4:
        image = convert_color_space(image, BGRA2RGBA)
    return image


def show_image(image, name='image', wait=True):
    """Shows RGB image in an external window.

    # Arguments
        image: Numpy array
        name: String indicating the window name.
        wait: Boolean. If ''True'' window stays open until user presses a key.
            If ''False'' windows closes immediately.
    """
    if image.dtype != np.uint8:
        raise ValueError('``image`` must be of type ``uint8``')
    # openCV default color space is BGR
    image = convert_color_space(image, RGB2BGR)
    cv2.imshow(name, image)
    if wait:
        while True:
            if cv2.waitKey(0) & 0xFF == ord('q'):
                break
        cv2.destroyAllWindows()


def warp_affine(image, matrix, fill_color=[0, 0, 0], size=None):
    """ Transforms `image` using an affine `matrix` transformation.

    # Arguments
        image: Numpy array.
        matrix: Numpy array of shape (2,3) indicating affine transformation.
        fill_color: List/tuple representing BGR use for filling empty space.
    """
    if size is not None:
        width, height = size
    else:
        height, width = image.shape[:2]
    return cv2.warpAffine(
        image, matrix, (width, height), borderValue=fill_color)


def write_image(filepath, image):
    """Writes an image inside ``filepath``. If ``filepath`` doesn't exist
        it makes a directory. If ``image`` has three channels the image is
        converted into BGR and then written. This is done such that this
        function compatible with ``load_image``.

    # Arguments
        filepath: String with image path. It should include postfix e.g. .png
        image: Numpy array.
    """
    directory_name = os.path.dirname(filepath)
    if (not os.path.exists(directory_name) and (len(directory_name) > 0)):
        os.makedirs(directory_name)
    if image.shape[-1] == 3:
        image = convert_color_space(image, RGB2BGR)
    return cv2.imwrite(filepath, image)


def gaussian_image_blur(image, kernel_size=(5, 5)):
    """Applies Gaussian blur to an image.

    # Arguments
        image: Numpy array of shape ''(H, W, 4)''.
        kernel_size: List of two ints e.g. ''(5, 5)''.

    # Returns
        Numpy array
    """
    return cv2.GaussianBlur(image, kernel_size, 0)


def median_image_blur(image, apperture=5):
    """Applies median blur to an image.

    # Arguments
        image: Numpy array of shape ''(H, W, 3)''.
        apperture. Int.

    # Returns
        Numpy array.
    """
    return cv2.medianBlur(image, apperture)


def get_rotation_matrix(center, degrees, scale=1.0):
    """Returns a 2D rotation matrix.

    # Arguments
        center: List of two integer values.
        degrees: Float indicating the angle in degrees.

    # Returns
        Numpy array
    """
    return cv2.getRotationMatrix2D(center, degrees, scale)


def get_affine_transform(source_points, destination_points):
    '''
    Return the transformation matrix.

    # Arguments
        source_points: Numpy array.
        destination_points: Numpy array.

    # Returns
        Transformation matrix.
    '''
    return cv2.getAffineTransform(source_points, destination_points)



from tensorflow.keras.losses import Loss
import tensorflow as tf


def split_alpha_mask(RGBA_mask):
    """Splits alpha mask and RGB image.

    # Arguments
        RGBA_mask: Tensor [batch, H, W, 4]

    # Returns
        Color tensor [batch, H, W, 3] and alpha tensor [batch, H, W, 1]
    """
    color_mask = RGBA_mask[:, :, :, 0:3]
    alpha_mask = RGBA_mask[:, :, :, 3:4]
    return color_mask, alpha_mask


def split_error_mask(RGBE_mask):
    """Splits error mask and RGB image.

    # Arguments
        RGBA_mask: Tensor [batch, H, W, 4]

    # Returns
        Color tensor [batch, H, W, 3] and error tensor [batch, H, W, 1]

    """
    color_mask = RGBE_mask[:, :, :, 0:3]
    error_mask = RGBE_mask[:, :, :, 3:4]
    return color_mask, error_mask


def compute_foreground_loss(RGB_true, RGB_pred, alpha_mask):
    """Computes foreground reconstruction L1 loss by using only positive alpha
        mask values.

    # Arguments
        RGB_true: Tensor [batch, H, W, 3]. True RGB label values.
        RGB_pred: Tensor [batch, H, W, 3]. Predicted RGB values.
        alpha_mask: Tensor [batch, H, W, 1]. True normalized alpha mask values.

    # Returns
        Tensor [batch, H, W, 3] with foreground loss values.
    """
    foreground_true = RGB_true * alpha_mask
    foreground_pred = RGB_pred * alpha_mask
    foreground_loss = tf.abs(foreground_true - foreground_pred)
    return foreground_loss


def compute_background_loss(RGB_true, RGB_pred, alpha_mask):
    """Computes the L1 reconstruction loss, weighting the inverted alpha
        mask values in the predicted RGB image by beta.

    # Arguments
        RGB_true: Tensor [batch, H, W, 3]. True RGB label values.
        RGB_pred: Tensor [batch, H, W, 3]. Predicted RGB values.
        alpha_mask: Tensor [batch, H, W, 1]. True normalized alpha mask values.

    # Returns
        Tensor [batch, H, W, 3] with background loss values.
    """
    background_true = RGB_true * (1.0 - alpha_mask)
    background_pred = RGB_pred * (1.0 - alpha_mask)
    background_loss = tf.abs(background_true - background_pred)
    return background_loss


def compute_weighted_reconstruction_loss(RGBA_true, RGB_pred, beta=3.0):
    """Computes the L1 reconstruction loss, weighting the positive alpha
        mask values in the predicted RGB image by beta.

    # Arguments
        RGBA_true: Tensor [batch, H, W, 4]. Color with alpha mask label values.
        RGB_pred: Tensor [batch, H, W, 3]. Predicted RGB values.
        beta: Float. Value used to multiple positive alpha mask values.

    # Returns
        Tensor [batch, H, W] with weighted reconstruction loss values.

    """
    RGB_true, alpha_mask = split_alpha_mask(RGBA_true)
    foreground_loss = compute_foreground_loss(RGB_true, RGB_pred, alpha_mask)
    background_loss = compute_background_loss(RGB_true, RGB_pred, alpha_mask)
    reconstruction_loss = (beta * foreground_loss) + background_loss
    return tf.reduce_mean(reconstruction_loss, axis=-1, keepdims=True)


def normalized_image_to_normalized_device_coordinates(image):
    """Map image value from [0, 1] -> [-1, 1].
    """
    return (image * 2.0) - 1.0


def normalized_device_coordinates_to_normalized_image(image):
    """Map image value from [0, 1] -> [-1, 1].
    """
    return (image + 1.0) / 2.0


def compute_weighted_reconstruction_loss_with_error(
        RGBA_true, RGBE_pred, beta=3.0):
    """Computes L1 reconstruction loss by multiplying positive alpha mask
        by beta.

    # Arguments
        RGBA_true: Tensor [batch, H, W, 4]. Color with alpha mask label values.
        RGBE_pred: Tensor [batch, H, W, 4]. Predicted RGB and error mask.
        beta: Float. Value used to multiple positive alpha mask values.

    # Returns
        Tensor [batch, H, W] with weighted reconstruction loss values.

    """
    RGB_pred, error_mask = split_error_mask(RGBE_pred)
    loss = compute_weighted_reconstruction_loss(RGBA_true, RGB_pred, beta)
    return loss


class WeightedReconstruction(Loss):
    """Computes L1 reconstruction loss by multiplying positive alpha mask
        by beta.

    # Arguments
        beta: Float. Value used to multiple positive alpha mask values.
        RGBA_true: Tensor [batch, H, W, 4]. Color with alpha mask label values.
        RGB_pred: Tensor [batch, H, W, 3]. Predicted RGB values.

    # Returns
        Tensor [batch, H, W] with weighted reconstruction loss values.

    """
    def __init__(self, beta=3.0):
        super(WeightedReconstruction, self).__init__()
        self.beta = beta

    def call(self, RGBA_true, RGB_pred):
        loss = compute_weighted_reconstruction_loss(
            RGBA_true, RGB_pred, self.beta)
        return loss


class WeightedReconstructionWithError(Loss):
    """Computes L1 reconstruction loss by multiplying positive alpha mask
        by beta.

    # Arguments
        RGBA_true: Tensor [batch, H, W, 4]. Color with alpha mask label values.
        RGBE_pred: Tensor [batch, H, W, 4]. Predicted RGB and error mask.
        beta: Float. Value used to multiple positive alpha mask values.

    # Returns
        Tensor [batch, H, W] with weighted reconstruction loss values.

    """
    def __init__(self, beta=3.0):
        super(WeightedReconstructionWithError, self).__init__()
        self.beta = beta

    def call(self, RGBA_true, RGBE_pred):
        reconstruction_loss = compute_weighted_reconstruction_loss_with_error(
            RGBA_true, RGBE_pred, self.beta)
        return reconstruction_loss



from .ssd300 import SSD300
from .ssd512 import SSD512
from .haar_cascade import HaarCascadeDetector
from .efficientdet import EFFICIENTDETD0
from .efficientdet import EFFICIENTDETD1
from .efficientdet import EFFICIENTDETD2
from .efficientdet import EFFICIENTDETD3
from .efficientdet import EFFICIENTDETD4
from .efficientdet import EFFICIENTDETD5
from .efficientdet import EFFICIENTDETD6
from .efficientdet import EFFICIENTDETD7



import numpy as np

from ..abstract import Processor
from ..backend.image import lincolor
from ..backend.image import draw_rectangle
from ..backend.image import put_text
from ..backend.image import draw_keypoint
from ..backend.image import draw_cube
from ..backend.image import GREEN
from ..backend.image import draw_random_polygon
from ..backend.image import draw_keypoints_link
from ..backend.image import draw_keypoints
from ..backend.image import draw_RGB_mask
from ..backend.image import draw_RGB_masks
from ..backend.keypoints import project_points3D
from ..backend.keypoints import build_cube_points3D
from ..backend.groups import quaternion_to_rotation_matrix
from ..backend.keypoints import project_to_image
from ..datasets import HUMAN_JOINT_CONFIG
from ..datasets import MINIMAL_HAND_CONFIG


class DrawBoxes2D(Processor):
    """Draws bounding boxes from Boxes2D messages.

    # Arguments
        class_names: List of strings.
        colors: List of lists containing the color values
        weighted: Boolean. If ``True`` the colors are weighted with the
            score of the bounding box.
        scale: Float. Scale of drawn text.
    """
    def __init__(self, class_names=None, colors=None,
                 weighted=False, scale=0.7, with_score=True):
        self.class_names = class_names
        self.colors = colors
        self.weighted = weighted
        self.with_score = with_score
        self.scale = scale

        if (self.class_names is not None and
                not isinstance(self.class_names, list)):
            raise TypeError("Class name should be of type 'List of strings'")

        if (self.colors is not None and
                not all(isinstance(color, list) for color in self.colors)):
            raise TypeError("Colors should be of type 'List of lists'")

        if self.colors is None:
            self.colors = lincolor(len(self.class_names))

        if self.class_names is not None:
            self.class_to_color = dict(zip(self.class_names, self.colors))
        else:
            self.class_to_color = {None: self.colors, '': self.colors}
        super(DrawBoxes2D, self).__init__()

    def call(self, image, boxes2D):
        for box2D in boxes2D:
            x_min, y_min, x_max, y_max = box2D.coordinates
            class_name = box2D.class_name
            color = self.class_to_color[class_name]
            if self.weighted:
                color = [int(channel * box2D.score) for channel in color]
            if self.with_score:
                text = '{:0.2f}, {}'.format(box2D.score, class_name)
            if not self.with_score:
                text = '{}'.format(class_name)
            put_text(image, text, (x_min, y_min - 10), self.scale, color, 1)
            draw_rectangle(image, (x_min, y_min), (x_max, y_max), color, 2)
        return image


class DrawKeypoints2D(Processor):
    """Draws keypoints into image.

    # Arguments
        num_keypoints: Int. Used initialize colors for each keypoint
        radius: Float. Approximate radius of the circle in pixel coordinates.
    """
    def __init__(self, num_keypoints, radius=3, normalized=False):
        super(DrawKeypoints2D, self).__init__()
        self.colors = lincolor(num_keypoints, normalized=normalized)
        self.radius = radius

    def call(self, image, keypoints):
        for keypoint_arg, keypoint in enumerate(keypoints):
            color = self.colors[keypoint_arg]
            draw_keypoint(image, keypoint.astype('int'), color, self.radius)
        return image


class DrawBoxes3D(Processor):
    def __init__(self, camera, class_to_dimensions,
                 color=GREEN, thickness=5, radius=2):
        """Draw boxes 3D of multiple objects

        # Arguments
            camera: Instance of ``paz.backend.camera.Camera''.
            class_to_dimensions: Dictionary that has as keys the
                class names and as value a list [model_height, model_width]
            thickness: Int. Thickness of 3D box
        """
        super(DrawBoxes3D, self).__init__()
        self.camera = camera
        self.class_to_dimensions = class_to_dimensions
        self.class_to_points = self._build_class_to_points(class_to_dimensions)
        self.color = color
        self.radius = radius
        self.thickness = thickness

    def _build_class_to_points(self, class_to_dimensions):
        class_to_points = {}
        for class_name, dimensions in self.class_to_dimensions.items():
            width, height, depth = dimensions
            points = build_cube_points3D(width, height, depth)
            class_to_points[class_name] = points
        return class_to_points

    def call(self, image, pose6D):
        points3D = self.class_to_points[pose6D.class_name]
        points2D = project_points3D(points3D, pose6D, self.camera)
        points2D = points2D.astype(np.int32)
        draw_cube(image, points2D, self.color, self.thickness, self.radius)
        return image


class DrawRandomPolygon(Processor):
    """ Adds occlusion to image

    # Arguments
        max_radius_scale: Maximum radius in scale with respect to image i.e.
                each vertex radius from the polygon is sampled
                from ``[0, max_radius_scale]``. This radius is later
                multiplied by the image dimensions.
    """
    def __init__(self, max_radius_scale=.5):
        super(DrawRandomPolygon, self).__init__()
        self.max_radius_scale = max_radius_scale

    def call(self, image):
        return draw_random_polygon(image)


def draw_pose6D(image, pose6D, points3D, intrinsics, thickness):
    """Draws cube in image by projecting points3D with intrinsics and pose6D.

    # Arguments
        image: Array (H, W).
        pose6D: paz.abstract.Pose6D instance.
        intrinsics: Array (3, 3). Camera intrinsics for projecting
            3D rays into 2D image.
        points3D: Array (num_points, 3).
        thickness: Positive integer indicating line thickness.

    # Returns
        Image array (H, W) with drawn inferences.
    """
    quaternion, translation = pose6D.quaternion, pose6D.translation
    rotation = quaternion_to_rotation_matrix(quaternion)
    points2D = project_to_image(rotation, translation, points3D, intrinsics)
    image = draw_cube(image, points2D.astype(np.int32), thickness=thickness)
    return image


class DrawPoses6D(Processor):
    """Draws multiple cubes in image by projecting points3D.

    # Arguments
        object_sizes: Array (3) indicating (x, y, z) sizes of object.
        camera_intrinsics: Array (3, 3).
            Camera intrinsics for projecting 3D rays into 2D image.
        thickness: Positive integer indicating line thickness.

    # Returns
        Image array (H, W) with drawn inferences.
    """
    def __init__(self, object_sizes, camera_intrinsics, thickness=2):
        self.points3D = build_cube_points3D(*object_sizes)
        self.intrinsics = camera_intrinsics
        self.thickness = thickness

    def call(self, image, poses6D):
        if poses6D is None:
            return image
        if not isinstance(poses6D, list):
            raise ValueError('Poses6D must be a list of Pose6D messages')
        for pose6D in poses6D:
            image = draw_pose6D(
                image, pose6D, self.points3D, self.intrinsics, self.thickness)
        return image


class DrawPose6D(Processor):
    """Draws a single cube in image by projecting points3D.

    # Arguments
        object_sizes: Array (3) indicating (x, y, z) sizes of object.
        camera_intrinsics: Array (3, 3).
            Camera intrinsics for projecting 3D rays into 2D image.
        thickness: Positive integer indicating line thickness.

    # Returns
        Image array (H, W) with drawn inferences.
    """
    def __init__(self, object_sizes, camera_intrinsics, thickness=2):
        self.points3D = build_cube_points3D(*object_sizes)
        self.intrinsics = camera_intrinsics
        self.thickness = thickness

    def call(self, image, pose6D):
        if pose6D is None:
            return image
        image = draw_pose6D(
            image, pose6D, self.points3D, self.intrinsics, self.thickness)
        return image


class DrawHumanSkeleton(Processor):
    """ Draw human pose skeleton on image.

    # Arguments
        images: Numpy array.
        grouped_joints: Joint locations of all the person model detected
                        in the image. List of numpy array.
        dataset: String.
        check_scores: Boolean. Flag to check score before drawing.

    # Returns
        A numpy array containing pose skeleton.
    """
    def __init__(self, dataset, check_scores, link_width=2, keypoint_radius=4):
        super(DrawHumanSkeleton, self).__init__()
        self.link_orders = HUMAN_JOINT_CONFIG[dataset]['part_orders']
        self.link_colors = HUMAN_JOINT_CONFIG[dataset]['part_color']
        self.link_args = HUMAN_JOINT_CONFIG[dataset]['part_arg']
        self.keypoint_colors = HUMAN_JOINT_CONFIG[dataset]['joint_color']
        self.check_scores = check_scores
        self.link_width = link_width
        self.keypoint_radius = keypoint_radius

    def call(self, image, grouped_joints):
        for one_person_joints in grouped_joints:
            image = draw_keypoints_link(
                image, one_person_joints, self.link_args, self.link_orders,
                self.link_colors, self.check_scores, self.link_width)
            image = draw_keypoints(image, one_person_joints,
                                   self.keypoint_colors, self.check_scores,
                                   self.keypoint_radius)
        return image


class DrawHandSkeleton(Processor):
    """ Draw hand pose skeleton on image.

    # Arguments
        image: Array (H, W, 3)
        keypoints: Array. All the joint locations detected by model
                        in the image.
    # Returns
        A numpy array containing pose skeleton.
    """
    def __init__(self, check_scores=False, link_width=2, keypoint_radius=4):
        super(DrawHandSkeleton, self).__init__()
        self.link_orders = MINIMAL_HAND_CONFIG['part_orders']
        self.link_colors = MINIMAL_HAND_CONFIG['part_color']
        self.link_args = MINIMAL_HAND_CONFIG['part_arg']
        self.keypoint_colors = MINIMAL_HAND_CONFIG['joint_color']
        self.check_scores = check_scores
        self.link_width = link_width
        self.keypoint_radius = keypoint_radius

    def call(self, image, keypoints):
        image = draw_keypoints_link(
            image, keypoints, self.link_args, self.link_orders,
            self.link_colors, self.check_scores, self.link_width)
        image = draw_keypoints(image, keypoints, self.keypoint_colors,
                               self.check_scores, self.keypoint_radius)
        return image


class DrawRGBMask(Processor):
    """Draws RGB mask by transforming points3D to RGB space and putting in
        them in their 2D coordinates (points2D)

    # Arguments
        object_sizes: Array (x_size, y_size, z_size)
    """
    def __init__(self, object_sizes):
        super(DrawRGBMask, self).__init__()
        self.object_sizes = object_sizes

    def call(self, image, points2D, points3D):
        image = draw_RGB_mask(image, points2D, points3D, self.object_sizes)
        return image


class DrawRGBMasks(Processor):
    """Draws RGB masks by transforming points3D to RGB space and putting in
        them in their 2D coordinates (points2D)

    # Arguments
        object_sizes: Array (x_size, y_size, z_size)
    """
    def __init__(self, object_sizes):
        super(DrawRGBMasks, self).__init__()
        self.object_sizes = object_sizes

    def call(self, image, points2D, points3D):
        return draw_RGB_masks(image, points2D, points3D, self.object_sizes)


class DrawText(Processor):
    """Draws text to image.

    # Arguments
        color: List. Color of text to
        thickness: Int. Thickness of text.
        scale: Int. Size scale for text.
        message: Str. Text to be added on the image.
        location: List/tuple of int. Pixel corordinte in image to add text.
    """
    def __init__(self, color=GREEN, thickness=2, scale=1):
        super(DrawText, self).__init__()
        self.color = color
        self.thickness = thickness
        self.scale = scale

    def call(self, image, message, location=(50, 50)):
        image = put_text(image, message, location, self.scale,
                         self.color, self.thickness)
        return image



from .pipelines import SSD512COCO
from .pipelines import SSD300VOC
from .pipelines import SSD512YCBVideo
from .pipelines import SSD300FAT
from .pipelines import DetectMiniXceptionFER
from .pipelines import MiniXceptionFER
from .pipelines import FaceKeypointNet2D32
from .pipelines import HeadPoseKeypointNet2D32
from .pipelines import HaarCascadeFrontalFace
from .pipelines import EFFICIENTDETD0COCO
from .pipelines import EFFICIENTDETD1COCO
from .pipelines import EFFICIENTDETD2COCO
from .pipelines import EFFICIENTDETD3COCO
from .pipelines import EFFICIENTDETD4COCO
from .pipelines import EFFICIENTDETD5COCO
from .pipelines import EFFICIENTDETD6COCO
from .pipelines import EFFICIENTDETD7COCO
from .pipelines import EFFICIENTDETD0VOC

from .pipelines import SinglePowerDrillPIX2POSE6D
from .pipelines import MultiPowerDrillPIX2POSE6D
from .pipelines import PIX2POSEPowerDrill
from .pipelines import PIX2YCBTools6D

from .pipelines import HigherHRNetHumanPose2D
from .pipelines import DetNetHandKeypoints
from .pipelines import MinimalHandPoseEstimation
from .pipelines import DetectMinimalHand
from .pipelines import ClassifyHandClosure
from .pipelines import SSD512MinimalHandPose



import tensorflow as tf

# same flags as in openCV
RGB2BGR = 4
BGR2RGB = 4
RGB2GRAY = 7
RGB2HSV = 41
HSV2RGB = 55


def cast_image(image, dtype):
    return tf.image.convert_image_dtype(image, dtype)


def load_image(filepath, num_channels=3):
    image = tf.io.read_file(filepath)
    image = tf.image.decode_image(image, num_channels, expand_animations=False)
    return image


def resize(image, size):
    return tf.image.resize(image, size)


def random_saturation(image, upper, lower):
    return tf.image.random_saturation(image, lower, upper)


def random_brightness(image, max_delta):
    return tf.image.random_brightness(image, max_delta)


def random_contrast(image, lower, upper):
    return tf.image.random_contrast(image, lower, upper)


def random_hue(image, max_delta):
    return tf.image.random_hue(image, max_delta)


def random_image_quality(image, lower, upper):
    return tf.image.random_jpeg_quality(image, lower, upper)


def _RGB_to_grayscale(image):
    return tf.image.rgb_to_grayscale(image)


def _RGB_to_HSV(image):
    return tf.image.rgb_to_hsv(image)


def _HSV_to_RGB(image):
    return tf.image.hsv_to_rgb(image)


def _reverse_channels(image):
    channels = tf.unstack(image, axis=-1)
    image = tf.stack([channels[2], channels[1], channels[0]], axis=-1)
    return image


def convert_color_space(image, flag):
    if flag == RGB2BGR:
        image = _reverse_channels(image)

    elif flag == BGR2RGB:
        image = _reverse_channels(image)

    elif flag == RGB2GRAY:
        image = _RGB_to_grayscale(image)

    elif flag == RGB2HSV:
        image = _RGB_to_HSV(image)

    elif flag == HSV2RGB:
        image = _HSV_to_RGB(image)

    elif flag == RGB2HSV:
        image = _RGB_to_HSV(image)

    else:
        raise ValueError('Invalid flag transformation:', flag)

    return image


def random_crop(image, size):
    return tf.image.random_crop(image, size)


def split_alpha_channel(image):
    if image.shape[-1] != 4:
        raise ValueError('Provided image does not contain alpha mask.')
    image, alpha_channel = tf.split(image, [3], -1)
    alpha_channel = alpha_channel / 255.0
    return image, alpha_channel


def alpha_blend(foreground, background, alpha_channel):
    return (alpha_channel * foreground) + ((1.0 - alpha_channel) * background)


def random_plain_background(image):
    """Adds random plain background to image using a normalized alpha channel
    # Arguments
        image: Float array-like with shape (H, W, 4).
        alpha_channel: Float array-like. Normalized alpha channel for blending.
    """
    image, alpha_channel = split_alpha_channel(image)
    random_color = tf.random.uniform([3], 0, 255)
    random_color = tf.reshape(random_color, [1, 1, 3])
    H, W = image.shape[:2]
    background = tf.tile(random_color, [H, W, 1])
    return alpha_blend(image, background, alpha_channel)


def random_cropped_background(image, background):
    image, alpha_channel = split_alpha_channel(image)
    background = random_crop(background, size=image.shape)
    return alpha_blend(image, background, alpha_channel)


def flip_left_right(image):
    return tf.image.flip_left_right(image)


def random_flip_left_right(image):
    if tf.random.uniform([1], 0, 2) == 1:
        image = flip_left_right(image)
    return image


def imagenet_preprocess_input(image, data_format=None, mode='torch'):
    image = tf.keras.applications.imagenet_utils.preprocess_input(image,
                                                                  data_format,
                                                                  mode)
    return image



import numpy as np

from ..abstract import Processor
from ..backend.boxes import to_one_hot
from ..backend.standard import append_values, predict


class ControlMap(Processor):
    """Controls which inputs are passed ''processor'' and the order of its
        outputs.

    # Arguments
        processor: Function e.g. a ''paz.processor''
        intro_indices: List of Ints.
        outro_indices: List of Ints.
        keep: ''None'' or dictionary. If ``None`` control maps operates
            without explicitly retaining an input. If dict it must contain
            as keys the input args to be kept and as values where they should
            be located at the end.
    """
    def __init__(self, processor, intro_indices=[0], outro_indices=[0],
                 keep=None):
        self.processor = processor
        if not isinstance(intro_indices, list):
            raise ValueError('``intro_indices`` must be a list')
        if not isinstance(outro_indices, list):
            raise ValueError('``outro_indices`` must be a list')
        self.intro_indices = intro_indices
        self.outro_indices = outro_indices
        name = '-'.join([self.__class__.__name__, self.processor.name])
        self.keep = keep
        super(ControlMap, self).__init__(name)

    def _select(self, inputs, indices):
        return [inputs[index] for index in indices]

    def _remove(self, inputs, indices):
        return [inputs[i] for i in range(len(inputs)) if i not in indices]

    def _split(self, inputs, indices):
        return self._select(inputs, indices), self._remove(inputs, indices)

    def _insert(self, args, extra_args, indices):
        [args.insert(index, arg) for index, arg in zip(indices, extra_args)]
        return args

    def call(self, *args):
        selected_args, remaining_args = self._split(args, self.intro_indices)
        processed_args = self.processor(*selected_args)
        if not isinstance(processed_args, tuple):
            processed_args = [processed_args]
        return_args = self._insert(
            remaining_args, processed_args, self.outro_indices)

        if self.keep is not None:
            keep_intro = list(self.keep.keys())
            keep_outro = list(self.keep.values())
            keep_args = self._select(args, keep_intro)
            return_args = self._insert(return_args, keep_args, keep_outro)

        return tuple(return_args)


class ExpandDomain(ControlMap):
    """Extends number of inputs a function can take applying the identity
    function to all new/extended inputs.
    e.g. For a given function f(x) = y. If g = ExtendInputs(f), we can
    now have g(x, x1, x2, ..., xn) = y, x1, x2, ..., xn.

    # Arguments
        processor: Function e.g. any procesor in ''paz.processors''.
    """
    def __init__(self, processor):
        super(ExpandDomain, self).__init__(processor)


class CopyDomain(Processor):
    """Copies ''intro_indices'' and places it ''outro_indices''.

    # Arguments
        intro_indices: List of Ints.
        outro_indices: List of Ints.
    """
    def __init__(self, intro_indices, outro_indices):
        super(CopyDomain, self).__init__()
        if not isinstance(intro_indices, list):
            raise ValueError('``intro_indices`` must be a list')
        if not isinstance(outro_indices, list):
            raise ValueError('``outro_indices`` must be a list')
        self.intro_indices = intro_indices
        self.outro_indices = outro_indices

    def _select(self, inputs, indices):
        return [inputs[index] for index in indices]

    def _insert(self, args, axes, values):
        [args.insert(axis, value) for axis, value in zip(axes, values)]
        return args

    def call(self, *args):
        selections = self._select(args, self.intro_indices)
        args = self._insert(list(args), self.outro_indices, selections)
        return tuple(args)


class UnpackDictionary(Processor):
    """Unpacks dictionary into a tuple.
    # Arguments
        order: List of strings containing the keys of the dictionary.
            The order of the list is the order in which the tuple
            would be ordered.
    """
    def __init__(self, order):
        if not isinstance(order, list):
            raise ValueError('``order`` must be a list')
        self.order = order
        super(UnpackDictionary, self).__init__()

    def call(self, kwargs):
        args = tuple([kwargs[name] for name in self.order])
        return args


class WrapOutput(Processor):
    """Wraps arguments in dictionary

    # Arguments
        keys: List of strings representing the keys used to wrap the inputs.
            The order of the list must correspond to the same order of
            inputs (''args'').
    """
    def __init__(self, keys):
        if not isinstance(keys, list):
            raise ValueError('``order`` must be a list')
        self.keys = keys
        super(WrapOutput, self).__init__()

    def call(self, *args):
        return dict(zip(self.keys, args))


class ExtendInputs(Processor):
    """Extends number of inputs a function can take applying the identity
    function to all new/extended inputs.
    e.g. For a given function f(x) = y. If g = ExtendInputs(f), we can
    now have g(x, x1, x2, ..., xn) = y, x1, x2, ..., xn.

    # Arguments
        processor: Function e.g. any procesor in ''paz.processors''.
    """
    def __init__(self, processor):
        self.processor = processor
        name = '-'.join([self.__class__.__name__, self.processor.name])
        super(ExtendInputs, self).__init__(name)

    def call(self, X, *args):
        return self.processor(X), args


class Concatenate(Processor):
    """Concatenates a list of arrays in given ''axis''.

    # Arguments
        axis: Int.
    """
    def __init__(self, axis):
        super(Concatenate, self)
        self.axis = axis

    def call(self, inputs):
        return np.concatenate(inputs, self.axis)


class SequenceWrapper(Processor):
    """Wraps arguments to directly use
    ''paz.abstract.ProcessingSequence'' or
    ''paz.abstract.GeneratingSequence''.

    # Arguments
        inputs_info: Dictionary containing an integer per key representing
            the argument to grab, and as value a dictionary containing the
            tensor name as key and the tensor shape of a single sample as value
            e.g. {0: {'input_image': [300, 300, 3]}, 1: {'depth': [300, 300]}}.
            The values given here are for the inputs of the model.
        labels_info: Dictionary containing an integer per key representing
            the argument to grab, and as value a dictionary containing the
            tensor name as key and the tensor shape of a single sample as value
            e.g. {2: {'classes': [10]}}.
            The values given here are for the labels of the model.
    """
    def __init__(self, inputs_info, labels_info):
        if not isinstance(inputs_info, dict):
            raise ValueError('``inputs_info`` must be a dictionary')
        self.inputs_info = inputs_info
        if not isinstance(labels_info, dict):
            raise ValueError('``inputs_info`` must be a dictionary')
        self.labels_info = labels_info
        self.inputs_name_to_shape = self._extract_name_to_shape(inputs_info)
        self.labels_name_to_shape = self._extract_name_to_shape(labels_info)
        self.ordered_input_names = self._extract_ordered_names(inputs_info)
        self.ordered_label_names = self._extract_ordered_names(labels_info)
        super(SequenceWrapper, self).__init__()

    def _extract_name_to_shape(self, info):
        name_to_shape = {}
        for values in info.values():
            for key, value in values.items():
                name_to_shape[key] = value
        return name_to_shape

    def _extract_ordered_names(self, info):
        arguments = list(info.keys())
        arguments.sort()
        names = []
        for argument in arguments:
            names.append(list(info[argument].keys())[0])
        return names

    def _wrap(self, args, info):
        wrap = {}
        for arg, name_to_shape in info.items():
            name = list(name_to_shape.keys())[0]
            wrap[name] = args[arg]
        return wrap

    def call(self, *args):
        inputs = self._wrap(args, self.inputs_info)
        labels = self._wrap(args, self.labels_info)
        return {'inputs': inputs, 'labels': labels}


class Predict(Processor):
    """Perform input preprocessing, model prediction and output postprocessing.

    # Arguments
        model: Class with a ''predict'' method e.g. a Keras model.
        preprocess: Function applied to given inputs.
        postprocess: Function applied to outputted predictions from model.
    """
    def __init__(self, model, preprocess=None, postprocess=None):
        super(Predict, self).__init__()
        self.model = model
        self.preprocess = preprocess
        self.postprocess = postprocess

    def call(self, x):
        return predict(x, self.model, self.preprocess, self.postprocess)


class ToClassName(Processor):
    def __init__(self, labels):
        super(ToClassName, self).__init__()
        self.labels = labels

    def call(self, x):
        return self.labels[np.argmax(x)]


class ExpandDims(Processor):
    """Expand dimension of given array.

    # Arguments
        axis: Int.
    """
    def __init__(self, axis):
        super(ExpandDims, self).__init__()
        self.axis = axis

    def call(self, x):
        return np.expand_dims(x, self.axis)


class SelectElement(Processor):
    """Selects element of input value.

    # Arguments
        index: Int. argument to select from ''inputs''.
    """
    def __init__(self, index):
        super(SelectElement, self).__init__()
        self.index = index

    def call(self, inputs):
        return inputs[self.index]


class BoxClassToOneHotVector(Processor):
    """Transform box data with class index to a one-hot encoded vector.

    # Arguments
        num_classes: Integer. Total number of classes.
    """
    def __init__(self, num_classes):
        self.num_classes = num_classes
        super(BoxClassToOneHotVector, self).__init__()

    def call(self, boxes):
        class_indices = boxes[:, 4].astype('int')
        one_hot_vectors = to_one_hot(class_indices, self.num_classes)
        one_hot_vectors = one_hot_vectors.reshape(-1, self.num_classes)
        boxes = np.hstack([boxes[:, :4], one_hot_vectors.astype('float')])
        return boxes


class Squeeze(Processor):
    """Wrap around numpy `squeeze` due to common use before model predict.
    # Arguments
        expand_dims: Int or list of Ints.
        topic: String.
    """
    def __init__(self, axis):
        super(Squeeze, self).__init__()
        self.axis = axis

    def call(self, x):
        return np.squeeze(x, axis=self.axis)


class Copy(Processor):
    """Copies value passed to function.
    """
    def __init__(self):
        super(Copy, self).__init__()

    def call(self, x):
        return x.copy()


class Lambda(object):
    """Applies a lambda function as a processor transformation.

    # Arguments
        function: Function.
    """

    def __init__(self, function):
        self.function = function

    def __call__(self, x):
        return self.function(x)


class StochasticProcessor(Processor):
    def __init__(self, probability=0.5, name=None):
        """Adds stochasticity to the user implemented ``call`` function

        # Arguments:
            probability: Probability of calling ``call`` function

        # Example:
        ```python
        class RandomAdd(StochasticProcessor):
        def __init__(self, probability=0.5):
            super(StochasticProcessor, self).__init__(probability)

        def call(self, x):
            return x + 1

        random_add = RandomAdd(probability=0.5)
        # value can be either 1.0 or 2.0
        value = random_add(1.0)
        ```
        """
        super(StochasticProcessor, self).__init__(name=name)
        self.probability = probability

    def call(self, X):
        raise NotImplementedError

    def __call__(self, X):
        if self.probability >= np.random.rand():
            return self.call(X)
        return X


class Stochastic(Processor):
    def __init__(self, function, probability=0.5, name=None):
        """Adds stochasticity to a given ``function``

        # Arguments:
            function: Callable object i.e. python function or
                ``paz.abstract.Processor``.
            probability: Probability of calling ``function``.

        # Example:
        ```python
        stochastic_add_one = Stochastic(lambda x: x + 1, 0.5)
        # value can be either 0.0 or 1.0
        value = random_add(0.0)
        ```
        """
        super(Stochastic, self).__init__(name=name)
        self.function = function
        self.probability = probability

    @property
    def probability(self):
        return self._probability

    @probability.setter
    def probability(self, probability):
        assert 0.0 <= probability <= 1.0, 'Probability must be between 0 and 1'
        self._probability = probability

    def call(self, X):
        if self.probability >= np.random.rand():
            return self.function(X)
        return X


class UnwrapDictionary(Processor):
    """Unwraps a dictionry into a list given the key order.
    """
    def __init__(self, keys):
        super(UnwrapDictionary, self).__init__()
        self.keys = keys

    def call(self, dictionary):
        return [dictionary[key] for key in self.keys]


class Scale(Processor):
    """Scales an input.
    """
    def __init__(self, scales):
        super(Scale, self).__init__()
        self.scales = scales

    def call(self, values):
        return self.scales * values


class AppendValues(Processor):
    """Append dictionary values to lists

    # Arguments
        keys: Keys to dictionary values
    """
    def __init__(self, keys):
        super(AppendValues, self).__init__()
        self.keys = keys

    def call(self, dictionary, lists):
        return append_values(dictionary, lists, self.keys)


class BooleanToTextMessage(Processor):
    """Convert a boolean to text message.
    # Arguments
        true_message: String. Message for true case.
        false_message: String. Message for false case.
        Flag: Boolean.

    # Returns
        message: String.
    """
    def __init__(self, true_message, false_message):
        super(BooleanToTextMessage, self).__init__()
        self.true_message = true_message
        self.false_message = false_message

    def call(self, flag):
        if flag:
            message = self.true_message
        else:
            message = self.false_message
        return message


class PrintTopics(Processor):
    """Prints topics
    # Arguments
        topics: List of keys to the inputted dictionary

    # Returns
        Returns same dictionary but outputs to terminal topic values.
    """
    def __init__(self, topics):
        super(PrintTopics, self).__init__()
        self.topics = topics

    def call(self, dictionary):
        [print(dictionary[topic]) for topic in self.topics]
        return dictionary



import os
import mmap

import numpy as np

from ..abstract import Loader


CLASS_DESCRIPTIONS_FILE = 'class-descriptions-boxable.csv'
BBOX_ANNOTATIONS_FILE = '{}-annotations-bbox.csv'


class OpenImages(Loader):
    """ Dataset loader for the OpenImagesV4 dataset.

    # Arguments
        path: String indicating full path to dataset
            e.g. /home/user/open_images/
        split: String determining the data split to load.
            e.g. `train`, `val` or `test`
        class_names: `all` or list. If list it should contain as elements
            the strings of the class names.

    """
    # TODO Allow selection of subset of class names.
    def __init__(self, path, split='train', class_names='all'):

        if split == 'val':
            split = 'validation'

        if split not in ['train', 'validation', 'test']:
            raise NameError('Invalid split name.')

        super(OpenImages, self).__init__(
            path, split, class_names, 'OpenImages')

        self.machine_to_human_name = dict()
        self.machine_to_arg = dict()
        self.load_class_names()
        self.class_distribution = dict()
        for class_name in self.class_names:
            self.class_distribution[class_name] = 0

    def load_class_names(self):
        classes_file = os.path.join(self.path, CLASS_DESCRIPTIONS_FILE)
        class_data = np.loadtxt(classes_file, delimiter=",", dtype=np.str)

        # class ID zero is background
        self.machine_to_arg['background'] = 0
        self.machine_to_human_name['background'] = 'background'
        class_names, class_arg = [], 1
        class_names.append('background')
        for machine_name, human_name in class_data:

            if self.class_names == 'all':
                self.machine_to_human_name[machine_name] = human_name
                self.machine_to_arg[machine_name] = class_arg
                class_names.append(human_name)
                class_arg = class_arg + 1

            elif human_name in self.class_names:
                self.machine_to_human_name[machine_name] = human_name
                self.machine_to_arg[machine_name] = class_arg
                class_names.append(human_name)
                class_arg = class_arg + 1

        self._class_names = class_names
        self._num_classes = len(self.machine_to_arg)
        print("Found {} {} classes".format(self.num_classes, self.split))

    def _get_num_lines(self, file_path):
        file_data = open(file_path, "r+")
        buf = mmap.mmap(file_data.fileno(), 0)
        lines = 0
        while buf.readline():
            lines = lines + 1
        return lines

    def load_data(self):

        data = dict()
        annotations_filepath = os.path.join(
            self.path, BBOX_ANNOTATIONS_FILE.format(self.split))
        # num_lines = self._get_num_lines(annotations_filepath)
        machine_names = self.machine_to_human_name.keys()
        # load file manually, line by line, in order to reduce memory usage
        with open(annotations_filepath, 'r') as annotations_file:
            # skip header
            annotations_file.readline()

            for line in annotations_file:
                row = line.split(",")

                image_filename = row[0] + ".jpg"
                x_min = float(row[4])
                x_max = float(row[5])
                y_min = float(row[6])
                y_max = float(row[7])

                machine_name = row[2]
                if machine_name not in machine_names:
                    continue

                human_name = self.machine_to_human_name[machine_name]

                absolute_image_path = os.path.join(
                    self.path, self.split, image_filename)

                if human_name in self.class_names:
                    class_arg = self.machine_to_arg[machine_name]

                    if absolute_image_path not in data:
                        data[absolute_image_path] = []

                    sample_data = [x_min, y_min, x_max, y_max, class_arg]
                    data[absolute_image_path].append(sample_data)
                    self.class_distribution[human_name] += 1

        formatted_data = []
        for image_path, ground_truth in data.items():
            sample = {'image': image_path, 'boxes': ground_truth}
            formatted_data.append(sample)

        msg = '{} split: loaded {} images with {} bounding box annotations'
        num_of_boxes = sum(self.class_distribution.values())
        print(msg.format(self.split, len(data), num_of_boxes))
        return formatted_data



import numpy as np
from paz.datasets import MANOHandJoints
from paz.backend.groups import to_affine_matrix
from paz.backend.groups import rotation_matrix_to_compact_axis_angle


def calculate_relative_angle(absolute_rotation, links_origin_transform,
                             parents=MANOHandJoints.parents):
    """Calculate the realtive joint rotation for the minimal hand joints.

    # Arguments
        absolute_angles : Array [num_joints, 4].
        Absolute joint angle rotation for the minimal hand joints in
        Euler representation.

    # Returns
        relative_angles: Array [num_joints, 3].
        Relative joint rotation of the minimal hand joints in compact
        axis angle representation.
    """
    relative_angles = np.zeros((len(absolute_rotation), 3))
    for angle_arg in range(len(absolute_rotation)):
        rotation = absolute_rotation[angle_arg]
        transform = to_affine_matrix(rotation, np.array([0, 0, 0]))
        inverted_transform = np.linalg.inv(transform)
        parent_arg = parents[angle_arg]
        if parent_arg is not None:
            link_transform = links_origin_transform[parent_arg]
            child_to_parent_transform = np.dot(inverted_transform,
                                               link_transform)
            child_to_parent_rotation = child_to_parent_transform[:3, :3]
            parent_to_child_rotation = np.linalg.inv(child_to_parent_rotation)
            parent_to_child_rotation = rotation_matrix_to_compact_axis_angle(
                parent_to_child_rotation)
            relative_angles[angle_arg] = parent_to_child_rotation
    return relative_angles


def reorder_relative_angles(relative_angles, root_angle, children,
                            root_joints=[1, 4, 7, 10, 13]):
    """Reorder the relative angles according to the kinematic chain

    # Arguments
        relative_angles: Array
        root_angle: Array. root joint angle for the minimal hand
        children: List, Indexes of the children in the kinematic chain.

    # Returns
        angles: Array. Reordered relative angles
    """
    if root_angle.shape == (3, 3):
        root_angle = rotation_matrix_to_compact_axis_angle(root_angle)
    children_angles = relative_angles[children[1:], :]
    children_angles = np.concatenate(
        [np.expand_dims(root_angle, 0), children_angles])

    # insest zero relative angle to the root joints
    angles = np.insert(children_angles, root_joints, np.array([0, 0, 0]), 0)
    return angles


def change_link_order(joints, config1_labels, config2_labels):
    """Map data from config1_labels to config2_labels.

    # Arguments
        joints: Array
        config1_labels: input joint configuration
        config2_labels: output joint configuration

    # Returns
        Array: joints maped to the config2_labels
    """
    mapped_joints = []
    for joint_arg in range(len(config2_labels)):
        joint_label = config2_labels[joint_arg]
        joint_index_in_config1_labels = config1_labels.index(joint_label)
        joint_in_config1_labels = joints[joint_index_in_config1_labels]
        mapped_joints.append(joint_in_config1_labels)
    mapped_joints = np.stack(mapped_joints, 0)
    return mapped_joints


def is_hand_open(relative_angles, joint_name_to_arg, thresh):
    """Check is the hand is open by calculating relative pip joint angle norm.

       [(theta * ex), (theta * ey), (theta * ez)] = compact axis angle
       ex, ey, ez = normalized_axis
                  _______________________________________________
       norm =    / (theta**2) * [(ex**2) + (ey**2) + (ez**2)]
               \/

       => norm is directly proportional to the theta if axis is normalized.
          If hand is open the relative angle of the pip joint will be less as
          compared to the closed hand.

    # Arguments
        relative_angle: Array
        joint_name_to_arg: Dictionary for the joints
        thresh: Float. Threshold value for theta

    # Returns
        Boolean: Hand is open or closed.
    """
    index_finger_pip_arg = joint_name_to_arg['index_finger_pip']
    theta_index_pip = np.linalg.norm(relative_angles[index_finger_pip_arg])

    middle_finger_pip_arg = joint_name_to_arg['middle_finger_pip']
    theta_middle_pip = np.linalg.norm(relative_angles[middle_finger_pip_arg])

    ring_finger_pip_arg = joint_name_to_arg['ring_finger_pip']
    theta_ring_pip = np.linalg.norm(relative_angles[ring_finger_pip_arg])

    pinky_finger_pip_arg = joint_name_to_arg['pinky_pip']
    theta_pinky_pip = np.linalg.norm(relative_angles[pinky_finger_pip_arg])

    if theta_index_pip > thresh and theta_middle_pip > thresh and \
            theta_ring_pip > thresh and theta_pinky_pip > thresh:
        return False
    else:
        return True



import os
import glob
import numpy as np

from tensorflow.keras.utils import Sequence, get_file

from ..backend.image import load_image, resize_image, make_mosaic
from ..abstract import Loader
from ..utils.documentation import docstring


def download(split):
    """Downloads omniglot dataset from original repository source.

    # Arguments:
        split: String indicating dataset split i.e. `train` or `test`.

    # Returns:
        filepath string to data split directory.
    """
    ROOT_URL = 'https://github.com/brendenlake/omniglot/blob/master/python/'
    split_to_name = {'train': 'images_background', 'test': 'images_evaluation'}
    filename = split_to_name[split]
    URL = ROOT_URL + filename + '.zip?raw=true'
    directory = 'paz/datasets/omniglot'
    filepath = get_file(None, URL, cache_subdir=directory, extract=True)
    filepath = os.path.join(os.path.dirname(filepath), filename)
    return filepath


def build_keyname(string):
    """Builds keynames in lower case and without parenthesis.

    # Arguments:
        string: Keyname string.

    # Returns
        String name for easy dictionary access.
    """
    string = os.path.basename(string)
    translations = {ord('('): None, ord(')'): None}
    return string.translate(translations).lower()


def enumerate_filenames(root_path):
    """Enumerates all file names inside given path.

    # Arguments
        root_path: String, path in which to search.

    # Returns
        list of sorted file names inside root path.
    """
    wildcard = os.path.join(root_path, '*')
    directories = glob.glob(wildcard)
    directories = sorted(directories)
    return directories


def load_shot(filepath, shape):
    """Loads images and preprocess it by resizing and normalizing it.

    # Arguments
        filepath: String indicating path to image.
        shape: List of integers indicating new shape (height, width).

    # Returns
        image as numpy array.
    """
    image = load_image(filepath, num_channels=1)
    image = resize_image(image, (shape))
    image = image / 255.0
    return image


def load_shots(shot_filepaths, shape):
    """Loads all images in character directory

    # Arguments:
        shot_filepaths: String. Filepath to character images.
        shape: List of integers indicating new shape (height, width).

    # Returns:
        Image array with all shots
    """
    shots = []
    for shot_filepath in shot_filepaths:
        shots.append(load_shot(shot_filepath, shape))
    return np.array(shots)


def load_characters(character_filepaths, shape):
    """Loads all characters in data directory.

    # Arguments
        character_filepaths: String indicating path to images.
        shape: List of integers indicating new shape (height, width).

    # Returns
        Dictionary with key name character name and value image array.
    """
    characters = {}
    for character_filepath in character_filepaths:
        character_name = build_keyname(character_filepath)
        shot_filepaths = enumerate_filenames(character_filepath)
        shots = load_shots(shot_filepaths, shape)
        characters[character_name] = shots
    return characters


def load(split='train', shape=(28, 28), flat=True):
    """Loads omniglot dataset for in between and within alphabet sampling.

    # Arguments
        split: String. Either `train` or `test`. Indicates which split to load.
        shape: List of two integers indicating resize shape `(H, W)`.
        flat: Boolean. If `True` the returned data dictionary is organized
            using each possible character as a class, with each key being a
            number having as value an image array.
            If `False` the returned data dictionary is organized using as keys
            the language names and as value another dictionary with keys being
            the character number, and as value the image array.
            This is to perform either sampling between alpahabet (`flat=True`)
            or to perform sampling within alphabet (`flat=False`).
            Usually, neural few-shot learning algorithms have been tested using
            in between alphabet sampling, but the original authors tested using
            the more challenging within alphabet sampling.

    # Returns
        dictionary with class names as keys and image numpy arrays as values.
    """
    filepath = download(split)
    language_filepaths = enumerate_filenames(filepath)
    languages = {}
    for language_filepath in language_filepaths:
        language_name = build_keyname(language_filepath)
        character_directories = enumerate_filenames(language_filepath)
        characters = load_characters(character_directories, shape)
        languages[language_name] = characters
    return flatten(languages) if flat else languages


def flatten(dataset):
    """Removes language hierarchy by having classes as each possible character.

    # Arguments:
        dataset: Dictionary with key names language name, and as value a
            dictionary with key names character names, and value an image array

    # Returns:
        Dictionary with key names as numbers and as values image arrays.
    """
    flat_dataset = {}
    flat_key = 0
    for language_name, language in dataset.items():
        for character_name, characters in language.items():
            flat_dataset[flat_key] = characters
            flat_key = flat_key + 1
    return flat_dataset


def sample_between_alphabet(RNG, dataset, num_ways, num_shots, num_tests=1):
    """Samples classification problems with flat dataset.
    Each sample is a meta learning problem with classes from all languages.

    # Arguments:
        RNG: Numpy random number generator.
        dataset: Dictionary.
        num_ways: Int. Number of classes for each meta learning episode.
        num_shots: Int. Number of train images used at each episode.
        num_tests: In. Number of test images at each episode.

    # Returns:
        Two lists. First list has `(train_images, train_labels)` and
        Second list has `(test_images, test_labels)`.
    """
    # dataset is flat, easier for sampling without replacement
    random_classes = RNG.choice(list(dataset.keys()), num_ways, replace=False)
    test_images, test_labels = [], []
    shot_images, shot_labels = [], []
    num_samples = num_shots + num_tests  # num_shots + 1
    for label, class_name in enumerate(random_classes):
        images = RNG.choice(dataset[class_name], num_samples, replace=False)
        labels = np.full(num_samples, label)
        shot_images.append(images[:num_shots])
        shot_labels.append(labels[:num_shots])
        test_images.append(images[num_shots:])
        test_labels.append(labels[num_shots:])
    shot_images = np.concatenate(shot_images, axis=0)
    shot_labels = np.concatenate(shot_labels, axis=0)
    test_images = np.concatenate(test_images, axis=0)
    test_labels = np.concatenate(test_labels, axis=0)
    return (shot_images, shot_labels), (test_images, test_labels)


def sample_within_alphabet(RNG, dataset, num_ways, num_shots, num_tests=1):
    """Samples classification problems with class hierarchical dataset.
    Each sample is a meta learning problem with classes from the same language.

    # Arguments:
        RNG: Numpy random number generator.
        dataset: Dictionary.
        num_ways: Int. Number of classes for each meta learning episode.
        num_shots: Int. Number of train images used at each episode.
        num_tests: In. Number of test images at each episode.

    # Returns:
        Two lists. First list has `(train_images, train_labels)` and
        Second list has `(test_images, test_labels)`.
    """
    alphabet_name = RNG.choice(list(dataset.keys()))
    alphabet = dataset[alphabet_name]
    reuse = True if num_ways > len(alphabet) else False  # FIX as 2019 Lake
    class_names = RNG.choice(list(alphabet.keys()), num_ways, reuse).tolist()
    test_images, test_labels = [], []
    shot_images, shot_labels = [], []
    num_samples = num_shots + num_tests  # num_shots + 1
    for label, class_name in enumerate(class_names):
        images = RNG.choice(alphabet[class_name], num_samples, replace=False)
        labels = np.full(num_samples, label)
        shot_images.append(images[:num_shots])
        shot_labels.append(labels[:num_shots])
        test_images.append(images[num_shots:])
        test_labels.append(labels[num_shots:])
    shot_images = np.concatenate(shot_images, axis=0)
    shot_labels = np.concatenate(shot_labels, axis=0)
    test_images = np.concatenate(test_images, axis=0)
    test_labels = np.concatenate(test_labels, axis=0)
    return (shot_images, shot_labels), (test_images, test_labels)


class Generator(Sequence):
    """Data generator for omniglot dataset with meta-learning episodes
    # Arguments
        sampler:
        num_classes: Int. Number of classes for each meta learning episode.
        num_support: Int. Number of train images used at each episode.
        num_queries: In. Number of test images at each episode.
        image_shape: List of integers indicating new shape (height, width).
        num_steps: Int. Number of samples per epoch.
    """
    def __init__(self, sampler, num_classes, num_support, num_queries,
                 image_shape, num_steps=2000):
        self.sampler = sampler
        self.support_shape = (num_classes, num_support, *image_shape)
        self.queries_shape = (num_classes, num_queries, *image_shape)
        self.num_steps = num_steps

    def __len__(self):
        return self.num_steps

    def __getitem__(self, idx):
        (support, support_labels), (queries, queries_labels) = self.sampler()
        support = np.reshape(support, self.support_shape)
        queries = np.reshape(queries, self.queries_shape)
        return {'support': support, 'queries': queries}, queries_labels


def remove_classes(RNG, data, num_classes):
    """Removes classes by randomly taking out keys from data dictionary.

    # Arguments:
        RNG: Numpy random number generator.
        data: Dictionary with keys as class names and values image arrays.

    # Returns:
        Dictionary with number of classes euqal to `num_classes`.
    """
    keys = RNG.choice(len(data.keys()), num_classes, replace=False)
    data = {key: data[key] for key in keys}
    return data


def split_data(data, validation_split):
    """Splits data keys into training and validation.

    # Arguments:
        data: Dictionary with keys as class names and values image arrays.
        validation_split: Float between `[0, 1]`. Porcentange of training
            data to be used for validation.

    # Returns:
        Two dictionaries with train and vlaidation data dictionaries.
    """
    keys = list(data.keys())
    num_train_keys = int(len(keys) * (1 - validation_split))
    train_keys = keys[:num_train_keys]
    valid_keys = keys[num_train_keys:]
    train_data = {key: data[key] for key in train_keys}
    valid_data = {key: data[key] for key in valid_keys}
    return train_data, valid_data


def plot_language(language):
    """Plots all characters in a language

    # Arguments:
        language: Dict with characters names as keys and image arrays as values

    # Returns:
        Image array with all characters.
    """
    characters = []
    for characters_name, images in language.items():
        images = np.expand_dims(images, axis=-1)
        characters.append(make_mosaic(images, (5, 4), 10))
    characters = np.array(characters)
    characters = make_mosaic(characters, (8, 7), 20)
    return characters


@docstring(load)
class Omniglot(Loader):
    def __init__(self, split, shape, flat=True):
        self.shape = shape
        self.flat = flat
        super(Omniglot, self).__init__(None, split, None, 'Omniglot')

    def load_data(self):
        return load(self.split, self.shape, self.flat)



import tensorflow.keras.backend as K
import tensorflow as tf
from tensorflow.keras.layers import Layer
from tensorflow.keras.initializers import Constant
import numpy as np


class Conv2DNormalization(Layer):
    """Normalization layer as described in ParseNet paper.

    # Arguments
        scale: Float determining how much to scale the features.
        axis: Integer specifying axis of image channels.

    # Returns
        Feature map tensor normalized with an L2 norm and then scaled.

    # References
        - [ParseNet: Looking Wider to
            See Better](https://arxiv.org/abs/1506.04579)
    """
    def __init__(self, scale, axis=3, **kwargs):
        self.scale = scale
        self.axis = axis
        super(Conv2DNormalization, self).__init__(**kwargs)

    def build(self, input_shape):
        self.gamma = self.add_weight(
            name='gamma', shape=(input_shape[self.axis]),
            initializer=Constant(self.scale), trainable=True)
        # super(Conv2DNormalization, self).build(input_shape)

    def output_shape(self, input_shape):
        return input_shape

    def call(self, x, mask=None):
        return self.gamma * K.l2_normalize(x, self.axis)


class SubtractScalar(Layer):
    """Subtracts scalar value to tensor.

    # Arguments
        constant: Float. Value to be subtracted to all tensor values.
    """
    def __init__(self, constant, **kwargs):
        self.constant = constant
        super(SubtractScalar, self).__init__(**kwargs)

    def build(self, input_shape):
        super(SubtractScalar, self).build(input_shape)

    def call(self, x):
        return x - self.constant

    def compute_output_shape(self, input_shape):
        return input_shape


class ExpectedValue2D(Layer):
    """Calculates the expected value along ''axes''.

    # Arguments
        axes: List of integers. Axes for which the expected value
            will be calculated.
    """
    def __init__(self, axes=[2, 3], **kwargs):
        self.axes = axes
        super(ExpectedValue2D, self).__init__(**kwargs)

    def build(self, input_shape):
        self.num_keypoints = input_shape[1]
        self.feature_map_size = input_shape[2]
        super(ExpectedValue2D, self).build(input_shape)

    def call(self, x):
        range_x, range_y = self.meshgrid(self.feature_map_size)
        expected_x = K.sum(x * range_x, axis=self.axes)
        expected_y = K.sum(x * range_y, axis=self.axes)
        keypoints_stack = K.stack([expected_x, expected_y], -1)
        keypoints = K.reshape(keypoints_stack, [-1, self.num_keypoints, 2])
        return keypoints

    def compute_output_shape(self, input_shape):
        return (input_shape[0], self.num_keypoints, 2)

    def meshgrid(self, feature_map_size):
        """ Returns a meshgrid ranging from [-1, 1] in x, y axes."""
        r = np.arange(0.5, feature_map_size, 1) / (feature_map_size / 2) - 1
        range_x, range_y = tf.meshgrid(r, -r)
        return K.cast(range_x, 'float32'), K.cast(range_y, 'float32')


class ExpectedDepth(Layer):
    """Calculates the expected depth along ''axes''.
    This layer takes two inputs. First input is a depth estimation tensor.
    Second input is a probability map of the keypoints.
    It multiplies both values and calculates the expected depth.

    # Arguments
        axes: List of integers. Axes for which the expected value
            will be calculated.
    """
    def __init__(self, axes=[2, 3], **kwargs):
        self.axes = axes
        super(ExpectedDepth, self).__init__(**kwargs)

    def build(self, input_shape):
        self.num_keypoints = input_shape[0][1]
        super(ExpectedDepth, self).build(input_shape)

    def call(self, x):
        z_volume, uv_volume = x
        z = K.sum(z_volume * uv_volume, axis=self.axes)
        z = K.expand_dims(z, -1)
        return z

    def compute_output_shape(self, input_shape):
        return (input_shape[0][0], self.num_keypoints, 1)



from setuptools import setup
from setuptools import find_packages
import paz


if __name__ == "__main__":

    setup(name='pypaz',
          version=paz.__version__,
          description='Perception for Autonomous Systems',
          long_description='Perception for Autonomous Systems',
          author='Octavio Arriaga',
          author_email='octavio.arriaga@dfki.de',
          url='https://github.com/oarriaga/paz/',
          license='MIT',
          classifiers=[
              'Intended Audience :: Science/Research',
              'Topic :: Scientific/Engineering :: Artificial Intelligence',
              'Topic :: Scientific/Engineering :: Image Recognition',
              'Programming Language :: Python :: 3',
              'License :: OSI Approved :: MIT License'
          ],
          install_requires=['opencv-python', 'tensorflow', 'numpy'],
          packages=find_packages())



import tensorflow as tf
from tensorflow.keras.losses import Loss


def compute_jaccard_score(y_true, y_pred, class_weights=1.0):
    """Computes the Jaccard score. The Jaccard score is the intersection
    over union of the predicted with respect to real masks.

    # Arguments
        y_true: Tensor of shape ``(batch, H, W, num_channels)``.
        y_pred: Tensor of shape ``(batch, H, W, num_channels)``.
        class_weights: Float or list of floats of shape ``(num_classes)``.

    # Returns
        Tensor of shape ``(batch)`` containing the F beta score per sample.
    """
    intersection = tf.reduce_sum(y_true * y_pred, axis=[1, 2])
    union = tf.reduce_sum(y_true + y_pred, axis=[1, 2]) - intersection
    jaccard_score = (intersection) / (union + 1e-5)
    return class_weights * jaccard_score


class JaccardLoss(Loss):
    """Computes the Jaccard loss. The Jaccard score is the intersection
    over union of the predicted with respect to real masks.

    # Arguments
        class_weights: Float or list of floats of shape ``(num_classes)``.
    """
    def __init__(self, class_weights=1.0):
        super(JaccardLoss, self).__init__()
        self.class_weights = class_weights

    def call(self, y_true, y_pred):
        return 1.0 - compute_jaccard_score(y_true, y_pred, self.class_weights)




import tensorflow as tf
from tensorflow.keras.losses import Loss


def compute_focal_loss(y_true, y_pred, gamma=2.0, alpha=0.25):
    """Computes the Focal loss. The Focal loss down weights
        properly classified examples.

    # Arguments
        y_true: Tensor of shape ``(batch, H, W, num_channels)``.
        y_pred: Tensor of shape ``(batch, H, W, num_channels)``.
        gamma: Float.
        alpha: Float.
        class_weights: Float or list of floats of shape ``(num_classes)``.

    # Returns
        Tensor of shape ``(batch)`` containing the F beta score per sample.
    """
    y_pred = tf.clip_by_value(y_pred, 1e-5, 1.0 - 1e-5)
    modulator = alpha * tf.math.pow(1 - y_pred, gamma)
    focal_loss = - modulator * y_true * tf.math.log(y_pred)
    return focal_loss


class FocalLoss(Loss):
    """Computes the Focal loss. The Focal loss down weights
        properly classified examples.

    # Arguments
        gamma: Float.
        alpha: Float.
        class_weights: Float or list of floats of shape ``(num_classes)``.
    """
    def __init__(self, gamma=2.0, alpha=0.25):
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.alpha = alpha

    def call(self, y_true, y_pred):
        return compute_focal_loss(y_true, y_pred, self.gamma, self.alpha)



from warnings import warn

import numpy as np

from .. import processors as pr
from ..abstract import SequentialProcessor, Processor
from ..backend.keypoints import translate_keypoints
from ..backend.keypoints import arguments_to_image_points2D
from ..backend.keypoints import normalize_keypoints2D
from ..backend.keypoints import denormalize_keypoints2D
from ..backend.keypoints import normalize_keypoints
from ..backend.keypoints import denormalize_keypoints
from ..backend.keypoints import compute_orientation_vector
from ..backend.image import get_scaling_factor
from ..backend.keypoints import standardize
from ..backend.keypoints import filter_keypoints2D
from ..backend.keypoints import destandardize
from ..backend.keypoints import merge_into_mean
from ..backend.keypoints import filter_keypoints3D
from ..backend.keypoints import initialize_translation, solve_least_squares
from ..backend.keypoints import get_bones_length, compute_reprojection_error
from ..backend.keypoints import compute_optimized_pose3D


class ProjectKeypoints(Processor):
    """Projects homogenous keypoints (4D) in the camera coordinates system into
        image coordinates using a projective transformation.

    # Arguments
        projector: Instance of ''paz.models.Project''.
        keypoints: Numpy array of shape ''(num_keypoints, 3)''
    """
    def __init__(self, projector, keypoints):
        self.projector = projector
        self.keypoints = keypoints
        super(ProjectKeypoints, self).__init__()

    def call(self, world_to_camera):
        keypoints = np.matmul(self.keypoints, world_to_camera.T)
        keypoints = np.expand_dims(keypoints, 0)
        keypoints = self.projector.project(keypoints)[0]
        return keypoints


class NormalizeKeypoints2D(Processor):
    """Transform keypoints in image-size coordinates to normalized coordinates.

    # Arguments
        image_size: List of two ints indicating ''(height, width)''
    """
    def __init__(self, image_size):
        self.image_size = image_size
        super(NormalizeKeypoints2D, self).__init__()

    def call(self, keypoints):
        height, width = self.image_size[0:2]
        keypoints = normalize_keypoints2D(keypoints, height, width)
        return keypoints


class DenormalizeKeypoints2D(Processor):
    """Transform normalized keypoints coordinates into image-size coordinates.

    # Arguments
        image_size: List of two floats having height and width of image.
    """
    def __init__(self):
        super(DenormalizeKeypoints2D, self).__init__()

    def call(self, keypoints, image):
        height, width = image.shape[0:2]
        keypoints = denormalize_keypoints2D(keypoints, height, width)
        return keypoints


class NormalizeKeypoints(Processor):
    """Transform keypoints in image-size coordinates to normalized coordinates.

    # Arguments
        image_size: List of two ints indicating ''(height, width)''
    """
    def __init__(self, image_size):
        self.image_size = image_size
        warn('DEPRECATED please use normalize_keypoints2D')
        super(NormalizeKeypoints, self).__init__()

    def call(self, keypoints):
        height, width = self.image_size[0:2]
        keypoints = normalize_keypoints(keypoints, height, width)
        return keypoints


class DenormalizeKeypoints(Processor):
    """Transform normalized keypoints coordinates into image-size coordinates.

    # Arguments
        image_size: List of two floats having height and width of image.
    """
    def __init__(self):
        warn('DEPRECATED please use denormalize_keypoints2D')
        super(DenormalizeKeypoints, self).__init__()

    def call(self, keypoints, image):
        height, width = image.shape[0:2]
        keypoints = denormalize_keypoints(keypoints, height, width)
        return keypoints


class RemoveKeypointsDepth(Processor):
    """Removes Z component from keypoints.
    """
    def __init__(self):
        super(RemoveKeypointsDepth, self).__init__()

    def call(self, keypoints):
        return keypoints[:, :2]


class PartitionKeypoints(Processor):
    """Partitions keypoints from shape [num_keypoints, 2] into a list of the
    form ((2), (2), ....) and length equal to num_of_keypoints.
    """
    def __init__(self):
        super(PartitionKeypoints, self).__init__()

    def call(self, keypoints):
        keypoints = np.vsplit(keypoints, len(keypoints))
        keypoints = [np.squeeze(keypoint) for keypoint in keypoints]
        partioned_keypoints = []
        for keypoint_arg, keypoint in enumerate(keypoints):
            partioned_keypoints.append(keypoint)
        return np.asarray(partioned_keypoints)


class ChangeKeypointsCoordinateSystem(Processor):
    """Changes ``keypoints`` 2D coordinate system using ``box2D`` coordinates
        to locate the new origin at the openCV image origin (top-left).
    """
    def __init__(self):
        super(ChangeKeypointsCoordinateSystem, self).__init__()

    def call(self, keypoints, box2D):
        x_min, y_min, x_max, y_max = box2D.coordinates
        keypoints[:, 0] = keypoints[:, 0] + x_min
        keypoints[:, 1] = keypoints[:, 1] + y_min
        return keypoints


class TranslateKeypoints(Processor):
    """Applies a translation to keypoints.
    The translation is a list of length two indicating the x, y values.
    """
    def __init__(self):
        super(TranslateKeypoints, self).__init__()

    def call(self, keypoints, translation):
        return translate_keypoints(keypoints, translation)


class ArgumentsToImageKeypoints2D(Processor):
    """Convert array arguments into UV coordinates.

              Image plane

           (0,0)-------->  (U)
             |
             |
             |
             v

            (V)

    # Arguments
        row_args: Array (num_rows).
        col_args: Array (num_cols).

    # Returns
        Array (num_cols, num_rows) representing points2D in UV space.

    # Notes
        Arguments are row args (V) and col args (U). Image points are in UV
            coordinates; thus, we concatenate them in that order
            i.e. [col_args, row_args]
    """
    def __init__(self):
        super(ArgumentsToImageKeypoints2D, self).__init__()

    def call(self, row_args, col_args):
        image_points2D = arguments_to_image_points2D(row_args, col_args)
        return image_points2D


class ScaleKeypoints(Processor):
    """Scale keypoints to input image shape.

    # Arguments
        keypoints: Array. Detected keypoints by the model
        image: Array. Input image.

    # Returns
        Scaled keypoints: Array. keypoints scaled to input image shape.
    """
    def __init__(self, scale=1, shape=(128, 128)):
        super(ScaleKeypoints, self).__init__()
        self.scale = scale
        self.shape = shape

    def call(self, keypoints, image):
        scale = get_scaling_factor(image, self.scale, self.shape)
        scaled_keypoints = keypoints * scale
        return np.array(scaled_keypoints, dtype=np.uint)


class ComputeOrientationVector(Processor):
    """Calculate the orientation of keypoints links with 3D keypoints.

    # Arguments
        keypoints: Array. 3D keypoints

    # Returns
        orientation: Array. Orientation of keypoint links
    """
    def __init__(self, parents):
        super(ComputeOrientationVector, self).__init__()
        self.parents = parents

    def call(self, keypoints):
        orientation = compute_orientation_vector(keypoints, self.parents)
        return orientation


class MergeKeypoints2D(Processor):
    def __init__(self, args_to_mean):
        """ Merges keypoints together then takes the mean of the keypoints

        # Arguments
            args_to_mean: keypoints indices

        # Returns
            Filtered keypoints2D
        """
        super(MergeKeypoints2D, self).__init__()
        self.args_to_mean = args_to_mean

    def call(self, keypoints2D):
        return merge_into_mean(keypoints2D, self.args_to_mean)


class FilterKeypoints2D(Processor):
    def __init__(self, args_to_mean, h36m_to_coco_joints2D):
        """ Filter keypoints2D

        # Arguments
            args_to_mean: keypoints indices
            h36m_to_coco_joints2D: h36m joints indices

        # Returns
            Filtered keypoints2D
        """
        super(FilterKeypoints2D, self).__init__()
        self.h36m_to_coco_joints2D = h36m_to_coco_joints2D
        self.args_to_mean = args_to_mean

    def call(self, keypoints2D):
        return filter_keypoints2D(keypoints2D, self.args_to_mean,
                                  self.h36m_to_coco_joints2D)


class StandardizeKeypoints2D(Processor):
    def __init__(self, data_mean2D, data_stdev2D):
        """ Standardize 2D keypoints

        # Arguments
            data_mean2D: mean 2D
            data_stdev2D: standard deviation 2D

        # Return
            standerized keypoints2D
        """
        self.mean = data_mean2D
        self.stdev = data_stdev2D
        super(StandardizeKeypoints2D, self).__init__()

    def call(self, keypoints2D):
        return standardize(keypoints2D, self.mean, self.stdev)


class DestandardizeKeypoints2D(Processor):
    def __init__(self, data_mean3D, data_stdev3D, dim_to_use):
        """ Destandardize 2D keypoints

        # Arguments
            data_mean3D: mean 3D
            data_stdev3D: standard deviation 3D
            dim_to_use: dimensions to use

        # Return
            detandardize 2D keypoints
        """
        self.mean = data_mean3D
        self.stdev = data_stdev3D
        self.valid = dim_to_use
        super(DestandardizeKeypoints2D, self).__init__()

    def call(self, keypoints2D):
        data = keypoints2D.reshape(-1, 48)
        rearanged_data = np.zeros((len(data), len(self.mean)),
                                  dtype=np.float32)
        rearanged_data[:, self.valid] = data
        destandardize_data = destandardize(rearanged_data, self.mean,
                                           self.stdev)
        return destandardize_data


class OptimizeHumanPose3D(Processor):
    """ Optimize human 3D pose

    #Arguments
        solver: library solver
        camera_intrinsics: camera intrinsic parameters

    #Returns
        keypoints3D, optimized keypoints3D
    """
    def __init__(self, args_to_joints3D, solver, camera_intrinsics):
        super(OptimizeHumanPose3D, self).__init__()
        self.args_to_joints3D = args_to_joints3D
        self.camera_intrinsics = camera_intrinsics
        self.solver = solver

    def call(self, keypoints3D, keypoints2D):
        joints3D = filter_keypoints3D(keypoints3D, self.args_to_joints3D)
        root2D = keypoints2D[:, :2]
        length2D, length3D = get_bones_length(keypoints2D, joints3D)
        ratio = length3D / length2D
        initial_joint_translation = initialize_translation(
            root2D, self.camera_intrinsics, ratio)
        joint_translation = solve_least_squares(
            self.solver, compute_reprojection_error, initial_joint_translation,
            joints3D, keypoints2D, self.camera_intrinsics)
        optimized_poses3D = compute_optimized_pose3D(
            keypoints3D, joint_translation, self.camera_intrinsics)
        return joints3D, optimized_poses3D



from ..backend.groups.quaternion import rotation_vector_to_quaternion


class Box2D(object):
    """Bounding box 2D coordinates with class label and score.

    # Properties
        coordinates: List of float/integers indicating the
            ``[x_min, y_min, x_max, y_max]`` coordinates.
        score: Float. Indicates the score of label associated to the box.
        class_name: String indicating the class label name of the object.

    # Methods
        contains()
    """
    def __init__(self, coordinates, score, class_name=None):
        x_min, y_min, x_max, y_max = coordinates
        self.coordinates = coordinates
        self.class_name = class_name
        self.score = score

    @property
    def coordinates(self):
        return self._coordinates

    @coordinates.setter
    def coordinates(self, coordinates):
        x_min, y_min, x_max, y_max = coordinates
        if x_min >= x_max:
            raise ValueError('Invalid coordinate input x_min >= x_max')
        if y_min >= y_max:
            raise ValueError('Invalid coordinate input y_min >= y_max')

        self._coordinates = coordinates

    @property
    def class_name(self):
        return self._class_name

    @class_name.setter
    def class_name(self, class_name):
        self._class_name = class_name

    @property
    def score(self):
        return self._score

    @score.setter
    def score(self, score):
        self._score = score

    @property
    def center(self):
        x_center = (self._coordinates[0] + self._coordinates[2]) / 2.0
        y_center = (self._coordinates[1] + self._coordinates[3]) / 2.0
        return x_center, y_center

    @property
    def width(self):
        return abs(self.coordinates[2] - self.coordinates[0])

    @property
    def height(self):
        return abs(self.coordinates[3] - self.coordinates[1])

    def __repr__(self):
        return "Box2D({}, {}, {}, {}, {}, {})".format(
            self.coordinates[0], self.coordinates[1],
            self.coordinates[2], self.coordinates[3],
            self.score, self.class_name)

    def contains(self, point):
        """Checks if point is inside bounding box.

        # Arguments
            point: Numpy array of size 2.

        # Returns
            Boolean. 'True' if 'point' is inside bounding box.
                'False' otherwise.
        """
        assert len(point) == 2
        x_min, y_min, x_max, y_max = self.coordinates
        inside_range_x = (point[0] >= x_min) and (point[0] <= x_max)
        inside_range_y = (point[1] >= y_min) and (point[1] <= y_max)
        return (inside_range_x and inside_range_y)


class Pose6D(object):
    """ Pose estimation results with 6D coordinates.

        # Properties
            quaternion: List of 4 floats indicating (w, x, y, z) components.
            translation: List of 3 floats indicating (x, y, z)
                translation components.
            class_name: String or ``None`` indicating the class label name of
                the object.

        # Class Methods
            from_rotation_vector: Instantiates a ``Pose6D`` object using a
                rotation and a translation vector.
    """
    def __init__(self, quaternion, translation, class_name=None):
        self.quaternion = quaternion
        self.translation = translation
        self.class_name = class_name

    @property
    def quaternion(self):
        return self._quaternion

    @quaternion.setter
    def quaternion(self, coordinates):
        self._quaternion = coordinates

    @property
    def translation(self):
        return self._translation

    @translation.setter
    def translation(self, coordinates):
        self._translation = coordinates

    @property
    def class_name(self):
        return self._class_name

    @class_name.setter
    def class_name(self, class_name):
        self._class_name = class_name

    @classmethod
    def from_rotation_vector(
            cls, rotation_vector, translation, class_name=None):
        quaternion = rotation_vector_to_quaternion(rotation_vector)
        pose6D = cls(quaternion, translation, class_name)
        pose6D.rotation_vector = rotation_vector
        return pose6D

    def __repr__(self):
        quaternion_message = ' Quaternion: ({}, {}, {}, {}) '.format(
            self.quaternion[0], self.quaternion[1],
            self.quaternion[2], self.quaternion[3])
        translation_message = ' Translation: ({}, {}, {}) '.format(
            self.translation[0], self.translation[1], self.translation[2])
        pose_message = ['Pose6D: ', quaternion_message, translation_message]
        pose_message = '\n \t'.join(pose_message)
        return pose_message


class Keypoint3D(object):
    def __init__(self, coordinates, class_name=None):
        coordinates = coordinates
        class_name = class_name

    @property
    def coordinates(self, coordinates):
        return self._coordinates

    @coordinates.setter
    def coordinates(self, coordinates):
        num_keypoints = len(coordinates)
        if num_keypoints != 3:
            raise ValueError('Invalid 3D Keypoint length:', num_keypoints)
        self._coordinates = coordinates

    def project():
        raise NotImplementedError

    def unproject():
        raise NotImplementedError
